"""
This is my first try at creating an iterated Bayes Net with PyMC objects. 
c0 ~ Normal(0, 1)
    c0's space is [-inf, inf], and the pdf is standard Normal
d1 ~ Uniform(c0-.5, c0+.5)
    d1's space is {0,1}, and the pdf is uniform with E(d1)=c0
c1 ~ Normal(d1, 1)
    c1's space is [-inf, inf], and the pdf is Normal with mean=d1 and std=1

for t = 2:N
    iterate definitions of d1 and c1
""" 

from pymc import *
import numpy as np

# Defining the root node
c_0 = Normal('c_0', 0, 1)

# Creating a dictionary of Nodes and adding c_0 to it
c_name = 'c_0'
Nodes = {c_name: c_0}

# Definining the number of "timesteps" in the net
T = 4

# Iterating on top of the root node to create an iterated Bayes Net
for t in range(1, T):
    d_name = 'd_%s' %t
    # Note that d_name references c_name, which is c_{t-1}
    Nodes[d_name] = Uniform(d_name, lower=Nodes[c_name]-.5, upper=Nodes[c_name]+.5)
    c_name = 'c_%s' %t
    # Note that c_name now becomes c_t'''
    Nodes[c_name] = Normal(c_name, mu=Nodes[d_name], tau=1)

# Creating a model from the Nodes dictionary
# Note that there are many ways to create a model, but using a dictionary makes
# attributes out of the nodes
M = Model(Nodes)

# Graphing the model, suppressing deterministic nodes
# Note: You should change the path to your own path
graph.dag(M, format='jpeg', name='iterated_net_dag', \
    path='/Users/jamesbono/libnfg/PyNFG/bin/', legend=True, \
    collapse_deterministics=True)

# Returning the set of nodes in the Model
M.nodes

# Take a random draw from the model
M.draw_from_prior()

# Get the log probability of the current values of the net
M.logp

# Draw a random value of c_1, given the current values of its parents
# (equivalent to Nodes['c_1'].random())
M.c_1.random()

# Get the log probability of the current value of c_1, given the current value of its parents
M.c_1.logp

# Return c_1's value (equivalent to Nodes['c_1'].value)
M.c_1.value

# Return c_1's children (equivalent to Nodes['c_1'].children)
# Note that this gives you the deterministic children, i.e. c_1-.5 and c_1+.5
# to see the deterministic children in the graph, set collapse_deterministics=False
# in the graph.dag function above
M.c_1.children

# Return d_2's parents (equivalent to Nodes['d_1'].parents)
# Note that this gives you the deterministic parents, i.e. c_1-.5 and c_1+.5
M.d_2.parents

# defining a stochastic node whose parents have not yet been defined. This raises
# an error -- NameError: name 'daddy' is not defined
orphan = Normal( 'orphan_node', mu=daddy, tau=mommy )

