function [CPT_ScottAttL0] = GenCPT_ScottAttL0(thetaA, params, BinsOrSet,T)
% This function is used to generate the CPT for Scott's attacker's policy
% Inputs: Parameter thetaA representing the attacker's characteristics
%         Global parameters
%         Bins for V2
% Output: CPT_ScottAttL0
    
    % Initialize CPT_ScottAttL0. mNew depends on V2 and mPrev
    CPT_ScottAttL0 = zeros(BinsOrSet.N_m, BinsOrSet.N_V2, BinsOrSet.N_m);
    
    if ( ~exist('T') )
        T = 30000; % Number of times we repeat for each combination of inputs
    end
    
    for j = 1:BinsOrSet.N_m % for each possible previous m
        mPrev = BinsOrSet.mSet(j);
        
        for i = 1:BinsOrSet.N_V2 % for each bin of V3
            % Generate V2 uniformly from the bin
            if (i == 1) % Apply half-Gussian distribution. Only sample the left to the mean 
                mu = BinsOrSet.V2n3bins(2); % mean = upperbound of the lowest bin
                sigma = (BinsOrSet.V2n3bins(3)-BinsOrSet.V2n3bins(2))/sqrt(1-2/pi); % Stdev = range of right neighboring bin divided by a difference factor
                Ran = mu + sigma.*randn(T,1); % Generate full-normally distributed numbers
                V2 = BinsOrSet.V2n3bins(2) - abs(BinsOrSet.V2n3bins(2) - Ran); % Convert full-normal to half-normal
            elseif (i == BinsOrSet.N_V3) % Apply half-Guassian distribution. Only sample the right to the mean. 
                mu = BinsOrSet.V2n3bins(i); % mean = lowerbound of the highest bin
                sigma = (BinsOrSet.V2n3bins(i) - BinsOrSet.V2n3bins(i-1))/sqrt(1-2/pi); % Stdev = range of the left neighboring bin divided by a difference factor
                Ran = mu + sigma.*randn(T,1); % Generate full-normally distributed numbers
                V2 = BinsOrSet.V2n3bins(i) + abs(BinsOrSet.V2n3bins(i) - Ran); % Convert full-normal to half-normal
            else % Sample uniformly in the medium three bins
                V2 = BinsOrSet.V2n3bins(i) + (BinsOrSet.V2n3bins(i+1)...
                    - BinsOrSet.V2n3bins(i)).*rand(T,1);
            end
            
            Num = zeros(BinsOrSet.N_m, 1); % Initialize the tracker
            
            [mStar, VStar] = Find_star(V2, mPrev,T, params, BinsOrSet); % Find stars 
            
            mNew = (VStar > thetaA).*mStar + (VStar <= thetaA).*...
                (mPrev*ones(T,1) - sign(V2 - params.V0));
            
            mNew = min(mNew, params.M);
            mNew = max(mNew, -params.M); % Avoid m beyond the range
            
            [~,p] = histc(mNew, BinsOrSet.mSet); % Find the indices in mSet
            for s = 1:BinsOrSet.N_m % for each m value
                Num(s,1) = sum(p==s); % Count the number that this value is reached
            end
            CPT_ScottAttL0(:,i,j) = Num(:,1)/T; % Conditional probabilities
        end
    end
end

function [mStar, VStar] = Find_star(V2,mPrev,T,params, BinsOrSet)
% This function is used to find the star quantities as in Scott's write-up
% Inputs: V2; previous m; global parameters
% Outputs: mStar; VStar

    q3max = params.p3max; % q3 can vary from -p3max to p3max
    
    % Replicate the arrays for the purpose of vectorization
    Rep_mSet = repmat(BinsOrSet.mSet, T, 1);
    Rep_V2 = repmat(V2, 1, BinsOrSet.N_m);
    
    % According to Section 3.5 in Scott's writeup
    arg = abs(Rep_V2+(Rep_mSet-mPrev)*q3max*params.x2/params.M/params.V0-params.V0);
    [~,ind]= max(arg,[],2); % the index of the largest value in every row
    mStar = (BinsOrSet.mSet(ind))';
    VStar = abs(max(arg,[],2));
    
    
end
    
            
            
            
            
            
            
            
            
            
            
            
            
        