%Class for simulation
classdef Simulation
    properties (SetAccess = private)
        PGS;
        numSteps;
        memList;
        initFcn;
        postExecFcn;
        playerList;
        params;
    end % properties
    
   methods
        % Constructor
        function SIM = Simulation( initFcn, nSteps, memList, params )
            SIM.initFcn = initFcn;
            SIM.numSteps = nSteps;
            SIM.memList = memList;
            SIM.params = params;
            SIM.postExecFcn = [];
        end % Constructor
        
        % Simulate the entire episode
        function [SIM] = SetPlayerList(SIM, playerList)
            SIM.playerList = playerList;
        end
        
        % Get number of time steps
        function [numSteps] = GetNumSteps(SIM)
            numSteps = SIM.numSteps;
        end
        
        % Simulate the entire episode
        function [SIM, mems, moves, rewards] = Init(SIM)
            SIM.PGS = feval( SIM.initFcn, SIM.params );
            
            %Initialize powergrid system
            SIM.PGS = SIM.PGS.simulate();
            
            assert( length(SIM.memList) == length(SIM.playerList) );
            
            %Initialize memory and player behavior
            for pi = 1:length(SIM.playerList)
                mems{pi} = SIM.memList(pi).Init(SIM.params);
                [SIM.playerList(pi),moves{pi}] = ...
                                SIM.playerList(pi).Init(SIM.params);
                rewards{pi} = SIM.playerList(pi).GetReward(SIM.PGS, ...
                                mems, moves, SIM.params);
            end
            
        end
        
        % Set post execution function: a user function that is executed
        % at the end of each time step.
        function [SIM] = SetPostExecFcn(SIM, postExecFcn)
            SIM.postExecFcn = postExecFcn;
        end
        
        % Simulate the entire episode
        function [trackPGS, trackMems, trackMoves, trackRewards] = Run(SIM)
            [SIM, mems, moves, rewards] = SIM.Init();
            trackPGS{1} = SIM.PGS;
            trackMems{1} = mems;
            trackMoves{1} = moves;
            trackRewards{1} = rewards;
            
            for t = 1:SIM.numSteps
               [SIM, mems, moves, rewards] = SIM.Step(mems, moves);
               
               trackPGS{t+1} = SIM.PGS;
               trackMems{t+1} = mems;
               trackMoves{t+1} = moves;
               trackRewards{t+1} = rewards;
            end
        end
        
        % Take a step
        function [SIM, mems, moves, rewards] = Step(SIM, mem_prevs, move_prevs)
            %Sample powergrid system
            SIM.PGS = SIM.PGS.simulate();
            
            assert( length(SIM.playerList) == length(SIM.memList) );
            
            %Loop through to update memories
            for pi = 1:length(SIM.playerList)
                mems{pi} = SIM.memList(pi).Sample(SIM.PGS, mem_prevs{pi}, ...
                                move_prevs{pi}, SIM.params);
            end
            
            %Loop through to update player moves
            for pi = 1:length(SIM.playerList)
                moves{pi} = SIM.playerList(pi).Play( mems{pi}, SIM.params );
            end
            
            %Loop through to update rewards
            for pi = 1:length(SIM.playerList)
                rewards{pi} = SIM.playerList(pi).GetReward(SIM.PGS, mems,...
                        moves, SIM.params);
            end
            
            %Execute the moves
            for pi = 1:length(SIM.playerList)
                SIM.PGS = SIM.playerList(pi).ApplyMove( SIM.PGS, moves{pi},...
                            SIM.params);
            end
            
            %Execute post user function, if one is defined
            if ( isa(SIM.postExecFcn, 'function_handle') )
                SIM.PGS = feval( SIM.postExecFcn, SIM.PGS, SIM.params );
            end
        end
        
        % Simulate the entire episode
        function [AvgRewardList, AvgRewardSEMList] = AverageReward(SIM, nIters)
            nPlayers = length(SIM.playerList);
            
            epsAvgReward = zeros( nPlayers, nIters );
            for ii = 1:nIters
                [~, ~, ~, rewards] = Run(SIM);
                
                epsSum = zeros(nPlayers,1);
                nSteps = length(rewards);
                for si = 1:nSteps
                    epsSum = epsSum + cell2mat(rewards{si})';
                end
                epsAvgReward(:, ii) = epsSum ./ nSteps;
            end
            AvgRewardList = mean(epsAvgReward,2);
            AvgRewardSEMList = sqrt(var(epsAvgReward,0,2)./nIters);
        end
   end % methods 
end % classdef