
// *************************************************************************
//	LibNFG: Net-form game library
//
//	Author:		Ritchie Lee
//
//  Copyright 2011 United States Government as represented by the Administrator 
//  of the National Aeronautics and Space Administration.  All Rights Reserved.  
//  Licensed under the Apache License, Version 2.0 (the "License"); you may not 
//  use this file except in compliance with the License.  You may obtain a copy 
//  of the License at http://www.apache.org/licenses/LICENSE-2.0
//  Unless required by applicable law or agreed to in writing, software 
//  distributed under the License is distributed on an "AS IS" BASIS, WITHOUT 
//  WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the 
//  License for the specific language governing permissions and limitations 
//  under the License.
//
//	Revision Log
//
//	Username	Date		Description
//	-------------------------------------------------------
//	rlee		20110328	Created
//
// *************************************************************************

#ifndef INCLUDED_ALG_LEVELKRELAXED_H
#define INCLUDED_ALG_LEVELKRELAXED_H

#include <boost/function.hpp>
#include "bayesnet.h"

using namespace boost;

///	References: 
//	[Lee11] Lee & Wolpert "Game Theoretic Modeling of Pilot Behavior 
//						during Mid-Air Encounters " (2011).
//	[Koller09] Koller, D., Friedman, N. "Probabilistic graphical models: 
//						principles and techniques." MIT Press (2009)
class alg_LevelKRelaxed
{
public:

	//	Variants of the Level-K algorithm.  
	//	RELAXED - resample bnet each time.
	//	D-RELAXED - reuse samples from the d-separated set.
	enum LEVELK_ALGO { D_RELAXED, RELAXED };
	//	Variants in the players response.
	//	Best response: always play the move giving the highest (estimated) expected utility.
	//	Logit response: probability of playing a move is proportional to 
	//		the exponential of its (estimated) expected utility
	enum LEVELK_RESPONSE { BEST, LOGIT };

	/**	Constructor */
	alg_LevelKRelaxed( void );

	/**	Destructor */
	virtual ~ alg_LevelKRelaxed();

	/// Variables

	//	Methods
	/**
	 * Set_params: Set parameters needed for level-K algorithms.
	 *	bnet - nfg object
	 *	mode - algorithm type (e.g. relaxed, d_relaxed)
	 *	M - number of own moves to sample in reasoning process.  Described in [Lee11].
	 *	MP - number of samples of everything else in the net in the reasoning process.  
	 *		See description of M' in [Lee11].
	 *	fo_Sample_level0_inst - a vector (of size bnet num vertices) with function objects at the
	 *		positions of decision nodes for sampling their level0 distributions.
	 *	fo_Sample_move_inst - a vector (of size bnet num vertices) with function objects at the
	 *		positions of decision nodes for sampling their satisficing distributions.
	 *	response - type of player's response to the vector of estimated utilities (e.g. BEST or LOGIT)
	 *	normalize - true = use normalized estimator [Koller09]
	 *		false = use unnormalized estimator
	 *	lambda - parameter used in LOGIT response which has the form exp( lambda * E(u) ).
	 *		may be interpreted as a rationality parameter since lambda = 0 -> random behavior
	 *		and lambda = infinity -> BEST response.
	 */
	void			Set_params ( BayesNet& bnet, const LEVELK_ALGO mode, const int M, const int MP, 
						const nfg_SampleFuncs& fo_Sample_level0_inst,
						const nfg_SampleFuncs& fo_Sample_move_inst,
						const LEVELK_RESPONSE response = BEST,
						const bool normalize = true, const double lambda = 1.0 );
	/**
	 * Set_player_proposal_funcs: Define proposal distribution functions needed for
	 *		for importance sampling of the net.
	 *	bnet - nfg object
	 *	propFuncInsts - vector of vector of sample functions for proposal distributions (one function
	 *		per node in net).  First list is according to node from the perspective you're sampling
	 *		(typically decision nodes).  Second list is according to
	 *		the node for which the proposal distribution applies (so usually chance nodes).
	 *	propProbFuncInsts - vector of Get probability functions for proposal distribution
	 *		one function per node in net).  Same structure as propFuncInsts.
	 */
	void			Set_player_proposal_funcs( BayesNet& bnet,
										const nfg_PropFuncInsts& propFuncInsts,
										const nfg_PropProbFuncInsts& propProbFuncInsts );
	/**
	 * LW_forward_sample: Returns a top-level sample of the net-form game with players at 
	 *		level K using likelihood-weighted sampling.  Assumes empty evidence set,
	 *		no masking, and no available partial sample.  Most users will call this 
	 *		function to execute the level-K algorithms.
	 *	bnet - nfg object
	 *	K - decision nodes play at this level.  Described in [Lee11].
	 *	returns: a weighted instantiation which represents a sample of the net.
	 */
	nfg_WtdInst		LW_forward_sample( BayesNet& bnet, const int K );
	/**
	 * LW_forward_sample_detailed: Returns a top-level sample of the net-form game with all options available
	 *		This function is mostly for internal calls.  Most users will use the abbreviated version.
	 *	bnet - nfg object
	 *	evidence - vector (of size bnet num vertices) that is contains a valid vector value
	 *		if that value is observed/evidence, empty if no evidence.
	 *	K - decision nodes play at this level.  Described in [Lee11].
	 *	mask - for partial sampling of the net.  The mask is a vector (of size bnet num vertices)
	 *		that indicates at the vertex id position whether the vertex should be sampled.  
	 *		true = sample node.  false = don't sample node.
	 *	partialWtdInst - for partial sampling of the net.  Used for passing partially sampled
	 *		instantiations.  Intended use: values not sampled (because of masking) are passed
	 *		as a partial sample so that the function will complete the rest of the samples
	 *		to create a complete instantiation.  The partial instantiation is needed when accessing for parents.
	 *	propFuncInst - inst of callback functions for sampling using proposal distribution.
	 *	propProbFuncInst - inst of callback functions for get_probability using proposal distribution.
	 *	returns: a weighted instantiation which represents a sample of the net.
	 */
	nfg_OptWtdInst		LW_forward_sample_detailed( BayesNet& bnet, const nfg_OptInst& evidence,
												const int K, const nfg_Mask& mask, 
												const nfg_OptWtdInst& partialWtdInst,
												const nfg_PropFuncInst& propFuncInst,
												const nfg_PropProbFuncInst& propProbFuncInst );
	

	//	Debug

private:

	/// Variables
	LEVELK_ALGO			m_mode;				// Set the player's strategy algorithm to use
	int					m_M;				// M = number of own strategies to consider
	int					m_MP;				// M' = number of scenarios to consider
	nfg_Masks			m_dSepMaskList;		// mask list containing d-sep masks rel to each v in D
	bool				m_normalize;		// true = use normalized finite variance estimator
	LEVELK_RESPONSE		m_response;			// Select response policy after probabilities have been calculated
	double				m_lambda;			// Rationality parameter for logit reponse
	vector<nfg_OptWtdInst*>		m_partialInstPtrs;			// Pointers to all previous partial inst samples
	nfg_SampleFuncs				m_fo_Sample_level0_inst;	// For sampling level 0 distributions
	nfg_SampleFuncs				m_fo_Sample_move_inst;		// call back functions for sampling movespace.
															// i.e., satisficing distributions
	nfg_PropFuncInsts			m_propFuncInsts;			// Callback functions for sampling proposal distribution
															// vector.  one per node in bnet.
	nfg_PropProbFuncInsts		m_propProbFuncInsts;		// Callback functions for evaluating probability
															// from proposal distribution.  
															// vector, one per node in bnet.

	/// Methods
	/**
	 * LevelK_d_relaxed: executes level-K d-relaxed strategies to sample a decision node
	 *	bnet - nfg object
	 *	id - vertex id of decision node
	 *	paVecs - parent vectors (these are observed by the node)
	 *	K - level K
	 *	returns: a sample of the node
	 */
	nfg_Vec			LevelK_d_relaxed ( BayesNet& bnet, const nfg_Vertex id, 
									const nfg_Vecs& paVecs,
									const int K );
	/**
	 * LevelK_relaxed: executes level-K relaxed strategies to sample a decision node
	 *	bnet - nfg object
	 *	id - vertex id of decision node
	 *	paVecs - parent vectors (these are observed by the node)
	 *	K - level K
	 *	returns: a sample of the node
	 */
	nfg_Vec			LevelK_relaxed ( BayesNet& bnet, const nfg_Vertex v, 
									const nfg_Vecs& paVecs,
									const int K );
	/**
	 * LW_forward_sample_node_prop: Sample node (e.g. CHANCE, DECISION) taking into account
	 *		evidence.
	 *	bnet - nfg object
	 *	evidence - a vector (of size bnet num vertices) where evidence/observed values
	 *		are inserted at vertex id.
	 *	id - the vertex id of the node to sample
	 *	paVec - parent vectors of the node to sample
	 *	K - if it is a decision node, use this value of level K
	 *	propFunc - callback function for sampling using proposal distribution 
	 *		(inputs already pre-loaded).
	 *	propProbFunc - callback function for get_probability using proposal distribution.
	 *		(inputs already pre-loaded, provide only outVec value).
	 *	returns: a sample of the node and its weighting (will be 1 if sampled, 
	 *		otherwise the probability_prop of the evidence occurring.
	 */
	nfg_WtdVec		LW_forward_sample_node_prop ( BayesNet& bnet, const nfg_OptInst& evidence,
												const nfg_Vertex id, 
												const nfg_Vecs& paVecs,
												const int K,
												const nfg_PropFuncDirect& propFunc,
												const nfg_PropProbFuncDirect& propProbFunc );

	//	Helper methods
	/**
	 * Get_mask_dsep_all: Get a complete list of d-separated sets, one set from each node.
	 *	bnet - nfg object
	 *	returns: a vector (size num vertices) of masks.  Each mask (true for in d-sep set)
	 *		identifies the d-separated set relative to each node.
	 */
	nfg_Masks		Get_mask_dsep_all ( BayesNet& bnet );
	/**
	 * Get_mask_dsep: Get the d-separated set relative to node v.
	 *	bnet - nfg object
	 *	returns: a mask representing the d-separated set (true for in d-sep set)
	 */
	nfg_Mask		Get_mask_dsep ( BayesNet& bnet, const nfg_Vertex& v );
	/**
	 * Get_mask_succ_all: Get a complete list of successor sets, one set from each node.
	 *	bnet - nfg object
	 *	returns: a vector (size num vertices) of masks.  Each mask (true for in succ set)
	 *		identifies the successor set relative to each node.
	 */
	nfg_Masks		Get_mask_succ_all ( BayesNet& bnet );
	/**
	 * Get_mask_succ: Get the successor set relative to node v.
	 *	bnet - nfg object
	 *	returns: a mask representing the successor set (true for in succ set)
	 */
	nfg_Mask		Get_mask_succ ( BayesNet& bnet, const nfg_Vertex& v );
	/**
	 * Get_parent_values: Get values of parents for a node v
	 *	bnet - nfg object
	 *	inst - could be an instantiation or partial instantiation containing the values of the parents of v
	 *	v - vertex id
	 *	returns: a vector (size num parents of v) of vectors containing the values of the parents
	 *		ordered by ascending parent vertex id.
	 */
	nfg_Vecs		Get_parent_values( BayesNet& bnet, const nfg_OptInst& inst, const nfg_Vertex& v );
	/**
	 * Get_neg_mask: Negates a given mask. i.e. each true becomes false, and each false becomes true
	 *	mask - mask to negate
	 *	returns: negated mask
	 */
	nfg_Mask		Get_neg_mask ( const nfg_Mask& mask );

};

#endif

