
// *************************************************************************
//	LibNFG: Net-form game library
//
//	Author:		Ritchie Lee
//
//  Copyright 2011 United States Government as represented by the Administrator 
//  of the National Aeronautics and Space Administration.  All Rights Reserved.  
//  Licensed under the Apache License, Version 2.0 (the "License"); you may not 
//  use this file except in compliance with the License.  You may obtain a copy 
//  of the License at http://www.apache.org/licenses/LICENSE-2.0
//  Unless required by applicable law or agreed to in writing, software 
//  distributed under the License is distributed on an "AS IS" BASIS, WITHOUT 
//  WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the 
//  License for the specific language governing permissions and limitations 
//  under the License.
//
//	Revision Log
//
//	Username	Date		Description
//	-------------------------------------------------------
//	rlee		20101025	Created
//	rlee		20110428	Added time extension representation
//	rlee		20110531	Added is_acyclic
//	rlee		20110816	Aggregated all vertex properties into one struct
//
// *************************************************************************

#include "bayesnet.h"

#include <string>
#include <iostream>
#include <boost/lexical_cast.hpp>
#include <boost/graph/topological_sort.hpp>

using namespace std;
using namespace boost;

// *************************************************************************
// *************************************************************************
BayesNet::BayesNet( const string bnetName, const unsigned int numVertices, 
						  const nfg_Edges& edgeArray, 
						  const nfg_Dim& vertexSizes,
						  const nfg_Spaces& vertexSpaces,
						  const nfg_Mask& vertexIsCts,
						  const nfg_eVertexTypes& vertexTypes,
						  const nfg_Mask& vertexIsBranch,
						  const nfg_Mask& vertexHasDno )
{
	//	Error-check input, sizes are consistent
	bassert ( numVertices > 0 );
	bassert ( vertexSizes.size() == numVertices );
	bassert ( vertexSpaces.size() == numVertices );
	bassert ( vertexIsCts.size() == numVertices );
	bassert ( vertexTypes.size() == numVertices );
	bassert ( vertexIsBranch.size() == numVertices );

	//	Straight assign simple variables
	m_bnetName			= bnetName;
	m_numVertices		= numVertices;
	m_numEdges			= edgeArray.size();

	//	Create a graph
	m_graph = nfg_Graph( edgeArray.begin(), edgeArray.end(), numVertices);

	//	Keep a finger on the properties
	m_vertexProps			= get ( vertex_t(), m_graph );

	//	Fill in node properties
	for ( unsigned int i = 0; i < m_numVertices; i++ )
	{
		m_vertexProps[i].m_id		= i;
		m_vertexProps[i].m_name		= lexical_cast<string>(i);
		m_vertexProps[i].m_isCts	= vertexIsCts.at(i);
		m_vertexProps[i].m_size		= vertexSizes.at(i);
		m_vertexProps[i].m_space	= vertexSpaces.at(i);
		m_vertexProps[i].m_type		= vertexTypes.at(i);
		m_vertexProps[i].m_isBranch = vertexIsBranch.at(i);
		m_vertexProps[i].m_hasDno	= vertexHasDno.at(i);
		m_vertexProps[i].mp_cpd		= NULL;
	}

	return;
}
// *************************************************************************
// *************************************************************************
BayesNet::~BayesNet ( void )
{
}
// *************************************************************************
// *************************************************************************
void BayesNet::Set_vertex_names ( const vector<string>& vertexNames )
{
	//	Loop through and assign each one to the vertex properties
	for ( unsigned int i = 0; i < m_numVertices; i++ )
		m_vertexProps[i].m_name.assign( vertexNames.at(i) );
}
// *************************************************************************
// *************************************************************************
void BayesNet::Set_cpd ( const nfg_Vertex id, cpd* p_cpd )
{
	// Make sure that the id is valid
	bassert ( id < m_numVertices );

	m_vertexProps[id].mp_cpd = p_cpd;
}
// *************************************************************************
// *************************************************************************
void BayesNet::Set_dno_function ( const nfg_Vertex id, const nfg_DnoFunc& dnoFunc )
{
	// Make sure that the id is valid
	bassert ( id < m_numVertices );
	// Make sure that the node has a dno value
	bassert ( Get_vertex_dno(id) );
	// Make sure that the dnoFunc is valid
	bassert ( dnoFunc );

	m_vertexProps[id].m_fo_dnoFunc = dnoFunc;
}
// *************************************************************************
// *************************************************************************
void BayesNet::Print_vertex_order ( void )
{
	cout << "Vertices: ";

	//	Loop through and print out to console each vertex name
	foreach ( nfg_Vertex vi, vertices(m_graph) )
	{
		cout << Get_vertex_name(vi) << ", ";
	}

	cout << endl;
}
// *************************************************************************
// *************************************************************************
bool BayesNet::Check ( void )
{
	bool overall = true;
	bool test = true;

	cout << "Running NFG checks:" << endl; 
	cout << "===================" << endl << endl;

	cout << "Check acyclicity...";
	test = Is_acyclic();
	cout << ( test? "PASSED" : "FAILED" ) << endl;
	overall &= test;

	cout << "Check topological order...";
	test = Is_in_topological_order();
	cout << ( test ? "PASSED" : "FAILED" ) << endl;
	overall &= test;

	cout << "Check branch node conditions...";
	test = Check_branch_node_conditions();
	cout << ( test ? "PASSED" : "FAILED" ) << endl;
	overall &= test;

	cout << "Check that all CPDs are defined...";
	test = Check_all_cpds_defined();
	cout << ( test ? "PASSED" : "FAILED" ) << endl;
	overall &= test;

	// TODO: Check other NFG conditions

	cout << "Test End." << endl << endl;

	return overall;
}
// *************************************************************************
// *************************************************************************
bool BayesNet::Check_all_cpds_defined ( void )
{
	foreach ( nfg_Vertex v, vertices(m_graph) )
	{
		// if any vertices are NULL (undefined), test fails immediately
		if( Get_cpd(v) == NULL )
			return false;
	}
	
	// if all vertices are defined, test passes
	return true;
}
// *************************************************************************
// *************************************************************************
bool BayesNet::Is_acyclic ( void )
{
	//	Check that all children of v have not been visited before v.

	//	Start off with a mask (of size num vertices) all false
	nfg_Mask mask( m_numVertices, false );

	//	for each node in the graph...
	foreach ( nfg_Vertex v, vertices(m_graph) )
	{
		//	Check its children...
		foreach ( nfg_Vertex vch, adjacent_vertices(v, m_graph) )
		{
			//	...to make sure all children have not been previously visited
			//	if even one child has been previously visited, it is
			//	not acyclic
			if ( mask.at(vch) )
				return false;
		}
		
		//	If all children have not been previously visited, mark this node as visited.
		mask.at(v) = true;
	}

	return true;
}
// *************************************************************************
// *************************************************************************
bool BayesNet::Is_in_topological_order ( void )
{
	//	Check that all parents of v are visited before v.

	//	Start off with a mask (of size num vertices) all false
	nfg_Mask mask( m_numVertices, false );

	//	for each node in the graph...
	foreach ( nfg_Vertex v, vertices(m_graph) )
	{
		//	Check its parents...
		foreach ( nfg_Vertex vpa, inv_adjacent_vertices(v, m_graph) )
		{
			//	...to make sure all parents have been previously visited
			//	if even one parent has not been previously visited, it is
			//	not in a topological order
			if ( !mask.at(vpa) )
				return false;
		}
		
		//	If all parents have been previously visited, mark this node as visited.
		mask.at(v) = true;
	}

	return true;
}
// *************************************************************************
// *************************************************************************
bool BayesNet::Check_branch_node_conditions ( void )
{
	foreach( nfg_Vertex vi, vertices(m_graph) )
	{
		// Check each branch node
		if ( Get_vertex_branch(vi) )
		{
			// If the branch node is continuous, automatically fail
			if ( Get_vertex_cts(vi) )
				return false;

			// Check that node is of size 1
			if ( Get_vertex_size(vi) != 1 )
				return false;
			
			// Check that number of children is equal to its space
			if ( Get_vertex_space(vi).at(0).size() != out_degree( vi, m_graph ) )
				return false;

			// Check that each child has a possible dno value
			foreach( nfg_Vertex ch, adjacent_vertices(vi, m_graph) )
			{
				if ( !Get_vertex_dno(ch) )
					return false;
			}
		}
	}

	return true;
}
// *************************************************************************
// *************************************************************************
nfg_Vertices BayesNet::Topological_sort ( void )
{
	nfg_Vertices vlist;
	
	//	Use Boost's built-in function for topological sort
	topological_sort( m_graph, back_inserter(vlist) );
	// the built-in function returns a reverse order, so reverse it back
	reverse ( vlist.begin(), vlist.end() );

	//	Print it out to screen...
	foreach( nfg_Vertex vi, vlist )
		cout << Get_vertex_name(vi) << "," ;
	cout << endl;

	//	and return the list
	return vlist;
}


