
// *************************************************************************
//	LibNFG: Net-form game library
//
//	Author:		Ritchie Lee
//
//  Copyright 2011 United States Government as represented by the Administrator 
//  of the National Aeronautics and Space Administration.  All Rights Reserved.  
//  Licensed under the Apache License, Version 2.0 (the "License"); you may not 
//  use this file except in compliance with the License.  You may obtain a copy 
//  of the License at http://www.apache.org/licenses/LICENSE-2.0
//  Unless required by applicable law or agreed to in writing, software 
//  distributed under the License is distributed on an "AS IS" BASIS, WITHOUT 
//  WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the 
//  License for the specific language governing permissions and limitations 
//  under the License.
//
//	Revision Log
//
//	Username	Date		Description
//	-------------------------------------------------------
//	rlee		20101025	Created
//	rlee		20110428	Added time extension representation
//	rlee		20110531	Added is_acyclic
//	rlee		20110816	Aggregated all vertex properties into one struct
//
// *************************************************************************

#ifndef INCLUDED_BAYESNET_H
#define INCLUDED_BAYESNET_H

#include <string>
#include <boost/graph/graph_traits.hpp>
#include <boost/graph/adjacency_list.hpp>
#include <boost/graph/adjacency_iterator.hpp>
#include <boost/graph/properties.hpp>
#include <boost/function.hpp>
#include <boost/optional.hpp>
#include "cpd.h"
#include "types.h"
#include "player.h"

using namespace std;
using namespace boost;

enum nfg_eVertexType { CHANCE, DECISION };			// vertex type

//	Declare vertex attributes
struct vertex_t									// vertex properties
{
	typedef boost::vertex_property_tag kind;
};

struct nfg_VertexProps
{
	// Vertex properties
	unsigned int					m_id;			// vertex ids
	string							m_name;			// vertex names
	bool							m_isCts;		// vertex is continuous
	unsigned int					m_size;			// vector size (number of dimensions)
	nfg_Space						m_space;		// discrete: vector of vector of valid values
														// continuous: vector of vector of min/max pairs
														// describing valid regions on the real line
	cpd*							mp_cpd;				// vertex generic cpd object
	nfg_eVertexType					m_type;				// vertex type
	bool							m_isBranch;			// whether vertex is branch
	bool							m_hasDno;			// vertex has a dno value
	nfg_DnoFunc						m_fo_dnoFunc;		// list of funcs for dno criteria
};

//	Vertex properties
typedef property<vertex_t, nfg_VertexProps>							nfg_VertexProperty;
																		
//	Graph properties
typedef adjacency_list<vecS, vecS, bidirectionalS, nfg_VertexProperty>	nfg_Graph;
typedef graph_traits<nfg_Graph>::vertex_descriptor						nfg_Vertex;
typedef vector<nfg_Vertex>												nfg_Vertices;
typedef std::pair<int, int>												nfg_Edge;
typedef vector<nfg_Edge>												nfg_Edges;
typedef vector<nfg_eVertexType>											nfg_eVertexTypes;

//	Property map
typedef property_map<nfg_Graph, vertex_t>::type						nfg_VertexPropsMap;

//	Graph iterators
typedef graph_traits<nfg_Graph>::vertex_iterator					nfg_VertexIter;
typedef pair<nfg_VertexIter,nfg_VertexIter>							nfg_VertexRange;
typedef graph_traits<nfg_Graph>::edge_iterator						nfg_EdgeIter;
typedef pair<nfg_EdgeIter,nfg_EdgeIter>								nfg_EdgeRange;
typedef graph_traits<nfg_Graph>::in_edge_iterator					nfg_InEdgeIter;
typedef	pair<nfg_InEdgeIter,nfg_InEdgeIter>							nfg_InEdgeRange;
typedef graph_traits<nfg_Graph>::out_edge_iterator					nfg_OutEdgeIter;
typedef pair<nfg_OutEdgeIter,nfg_OutEdgeIter>						nfg_OutEdgeRange;
typedef adjacency_iterator_generator<nfg_Graph>::type				nfg_ChIter;
typedef pair<nfg_ChIter,nfg_ChIter>									nfg_ChRange;
typedef inv_adjacency_iterator_generator<nfg_Graph>::type			nfg_PaIter;
typedef pair<nfg_PaIter,nfg_PaIter>									nfg_PaRange;


// Class object representation of NFG and all its attributes
class BayesNet
{
public:

	/**	Constructor */
	/**
	 * BayesNet: Constructs an NFG object
	 *	name - name of bnet
	 *	numVertices - number of vertices in the net
	 *	vertexSizes - the size of each vertex.  if continuous, size of vector.  
	 *		if discrete, number of states. Indexed by vertex id.
	 *	vertexSpaces - represents the space of possible values for each vertex
	 *	vertexIsCts - for each vertex indicates whether vertex is true = continuous,
	 *		false = discrete.  Indexed by vertex id.
	 *	vertexTypes - for each vertex indicates nfg_eVertexType (e.g. chance or decision)
	 *		Indexed by vertex id.
	 *	vertexIsBranch - for each vertex indicates whether node is a branch node.
	 *		true = branch. false = not branch. Index by vertex id.
	 *	pFuncs - partition function objects. Required for proper branching. Each 
	 *		branch node should have one to specify how to branch to children.
	 *		Indexed by vertex id.  Can leave blank if no branch nodes.
	 */
	BayesNet( const string name, const unsigned int numVertices,
							const nfg_Edges& edgeArray, 
							const nfg_Dim& vertexSizes,
							const nfg_Spaces& vertexSpaces,
							const nfg_Mask& vertexIsCts, 
							const nfg_eVertexTypes& vertexTypes,
							const nfg_Mask& vertexIsBranch,
							const nfg_Mask& vertexHasDno );
	/**
	 * BayesNet: Constructs an NFG object.  Default constructor.
	 *		This creates an empty object.  For developer use only.
	 */
	BayesNet(){};

	/**	Destructor */
	virtual ~ BayesNet();

	/// Variables

	
	///	Methods
	/**
	 * Get_graph: Returns an object reference to the actual graph for direct access calls
	 *		such as vertices(m_graph).
	 */
	const		nfg_Graph&		Get_graph ( void )	{ return m_graph;		};
	/**
	 * Get_bnet_name: Returns the name of the NFG.
	 */
	string		Get_bnet_name ( void )	{ return m_bnetName;	};
	/**
	 * Set_vertex_names: Sets the vertex names to NFG object
	 *	vertexNames - indexed by vertex id.  name of each vertex in the net.
	 */
	void		Set_vertex_names ( const vector<string>& vertexNames	);
	/**
	 * Get_vertex_name: Returns the name of vertex at id.
	 *	id - id of vertex of interest
	 *	returns: vertex name
	 */
	string		Get_vertex_name ( const nfg_Vertex id	) { return m_vertexProps[id].m_name; };
	/**
	 * Get_vertex_space: Returns the space of a vertex.
	 *	id - id of vertex of interest
	 *	returns: vertex space
	 */
	nfg_Space	Get_vertex_space ( const nfg_Vertex id ) { return m_vertexProps[id].m_space; };
	/**
	 * Get_vertex_type: Returns the type of the vertex at id.
	 *	id - id of vertex of interest
	 *	returns: vertex type
	 */
	nfg_eVertexType Get_vertex_type ( const nfg_Vertex id ) { return m_vertexProps[id].m_type; };
	/**
	 * Get_vertex_cts: Check whether vertex is continuous of discrete
	 *	id - id of vertex of interest
	 *	returns: true if vertex at id is continuous, false otherwise.
	 */
	bool		Get_vertex_cts ( const nfg_Vertex id ) { return m_vertexProps[id].m_isCts; };
	/**
	 * Get_vertex_branch: Check whether vertex is a branch node
	 *	id - id of vertex of interest
	 *	returns: true if vertex at id is a branch node, false otherwise.
	 */
	bool		Get_vertex_branch ( const nfg_Vertex id ) { return m_vertexProps[id].m_isBranch; };
	/**
	 * Get_vertex_size: Get size of vertex.
	 *	id - id of vertex of interest
	 *	returns: size of vector.
	 */
	unsigned int	Get_vertex_size ( const nfg_Vertex id ) { return m_vertexProps[id].m_size; };
	/**
	 * Get_num_vertices: Get number of vertices in NFG
	 *	returns: number of vertices in graph
	 */
	unsigned int	Get_num_vertices ( void ) { return m_numVertices; };
	/**
	 * Get_cpd: Get pointer to the conditional probability distribution generic object
	 *	id - id of vertex of interest
	 *	returns: a pointer to the cpd stored in vertex.
	 */
	cpd*		Get_cpd ( const nfg_Vertex id ) { return m_vertexProps[id].mp_cpd; };
	/**
	 * Set_cpd: Set pointer to the conditional probability distribution generic object
	 *	id - id of vertex of interest
	 *	p_cpd - pointer to cpd object
	 */
	void		Set_cpd ( const nfg_Vertex id, cpd* p_cpd );
	/**
	 * Get_vertex_dno: Get whether a vertex has a dno value
	 *	id - id of vertex of interest
	 *	returns: true if has dno, false otherwise
	 */
	bool	Get_vertex_dno ( const nfg_Vertex id ) { return m_vertexProps[id].m_hasDno; };
	/**
	 * Print_vertex_order: Prints out to screen the vertex names in order.
	 */
	void		Print_vertex_order ( void );
	/**
	 * Check_all_cpds_defined: Checks to make sure 
	 *	returns: true = is acyclic, false = otherwise
	 */
	bool		Check_all_cpds_defined ( void );
	/**
	 * Is_acyclic: Determines whether the graph is acyclic
	 *	returns: true = is acyclic, false = otherwise
	 */
	bool		Is_acyclic ( void );
	/**
	 * Is_in_topological_order: Determines whether current order is topological
	 *	returns: true = is in topological order, false = not in topological order
	 */
	bool		Is_in_topological_order ( void );
	/**
	 * Check_branch_node_conditions: Check to make sure all the conditions related to
	 *		branch nodes are satisfied (according to definition in Lee & Wolpert paper)
	 *	returns: true = satisfies all conditions, false = does not satisfy
	 */
	bool		Check_branch_node_conditions ( void );
	/**
	 * Topological_sort: suggests a topological order, use this if Is_in_topological_order 
	 *		returns false
	 *	returns: vertex list containing a possible order.
	 */
	nfg_Vertices Topological_sort( void );
	/**
	 * Add_player: Add a player to the list 
	 *	player - player object to be added
	 */
	void	Add_player ( const Player& player ) { m_players.push_back(player); };
	/**
	 * Get_player_list: Get the entire player list
	 *	returns: player list
	 */
	vector<Player>	Get_player_list ( void ) { return m_players; };
	/**
	 * Clear_player_list: Clear the player list
	 */
	void	Clear_player_list ( void ) { m_players.clear(); };
	/**
	 * Get_num_players: Get the number of players in the list
	 *	returns: player object to be added
	 */
	unsigned int	Get_num_players ( void ) { return m_players.size(); };
	/**
	 * Check: All-in-one error checking for NFG configuration
	 *	returns: true = is valid, false = is not valid
	 */
	bool		Check ( void );
	/**
	 * Set_dno_function: Set the dno function of a node that has dno value
	 *	id - id of vertex of interest
	 *	dnoFunc - function object of dno function (meta-rule determining 
	 *		whether a node should be dno)
	 */
	void		Set_dno_function ( const nfg_Vertex id, const nfg_DnoFunc& dnoFunc );
	/**
	 * Get_dno_func: Returns the dno function of a vertex
	 *	id - id of vertex of interest
	 *	returns: function object of dno function
	 */
	nfg_DnoFunc		Get_dno_func ( const nfg_Vertex id ) { return m_vertexProps[id].m_fo_dnoFunc; };
private:

	/// Variables
	nfg_Graph						m_graph;			// Core data object, from Boost Graph Library
	string							m_bnetName;			// Name of bayes net
	unsigned int					m_numVertices;		// Number of vertices
	unsigned int					m_numEdges;			// Number of edges
	nfg_VertexPropsMap				m_vertexProps;		// List of structs of vertex properties
	vector<Player>					m_players;			// List of players

	/// Methods

};

#endif

