
// *************************************************************************
//	LibNFG: Net-form game library
//
//	Author:		Ritchie Lee
//
//  Copyright 2011 United States Government as represented by the Administrator 
//  of the National Aeronautics and Space Administration.  All Rights Reserved.  
//  Licensed under the Apache License, Version 2.0 (the "License"); you may not 
//  use this file except in compliance with the License.  You may obtain a copy 
//  of the License at http://www.apache.org/licenses/LICENSE-2.0
//  Unless required by applicable law or agreed to in writing, software 
//  distributed under the License is distributed on an "AS IS" BASIS, WITHOUT 
//  WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the 
//  License for the specific language governing permissions and limitations 
//  under the License.
//
//	Revision Log
//
//	Username	Date		Description
//	-------------------------------------------------------
//	rlee		20101025	Created
//	rlee		20110816	Converted to inherited class
//	rlee		20111001	Added bias term
//
// *************************************************************************

#include "cpd_gaussian.h"

#include <boost/random.hpp>
#include <boost/math/distributions/normal.hpp>

using namespace nfg;
using namespace boost::math;

// *************************************************************************
// *************************************************************************
cpd_Gaussian::cpd_Gaussian( const nfg_Vec& stddev, const nfg_Vec& bias )
{
	//	For now, implements multivariate gaussian, but assumes a diagonal covariance matrix
	//	1 input vector, 1 output vector, 1 stddev vector.  All same size.

	bassert ( !stddev.empty() );
	
	//	Check to make sure all stddev's are positive and finite
	foreach ( double s, stddev )
	{
		bassert ( s == s );	// Check that s is finite
		bassert ( s >= 0 );	// Check that stddev is non-negative
	}

	m_stddev	= stddev;
	m_size		= m_stddev.size();

	// if no bias vector is given, no bias
	if( bias.empty() )
		m_bias = nfg_Vec( m_size, 0 );
	// if the sizes match, accept input bias vector
	else if( bias.size() == m_size )
		m_bias = bias;
	else
	{
		cout << "Error in cpd_Gaussian::cpd_Gaussian: Bias vector is invalid" << endl;
		bassert( false );
	}

}
// *************************************************************************
// *************************************************************************
cpd_Gaussian::~cpd_Gaussian ( void )
{
}
// *************************************************************************
// *************************************************************************
nfg_Vec cpd_Gaussian::Sample_node ( const nfg_Vecs& paVecs )
{
	//	assert only 1 input vector
	bassert ( paVecs.size() == 1 );
	
	//	assert sizes match
	bassert ( m_size == paVecs.at(0).size() );

	//	Use helper function call to generate a gaussian sample centered at 
	//	parent vector and with pre-specified member stddev.
	nfg_Vec outVec(m_size);
	for( unsigned int i = 0; i < paVecs.at(0).size(); i++ )
	{
		outVec.at(i) = Rand_gaussian ( paVecs.at(0).at(i) - m_bias.at(i), m_stddev.at(i) );
	}

	return outVec;
}
// *************************************************************************
// *************************************************************************
double cpd_Gaussian::Get_probability_prop ( const nfg_Vec& outVec, const nfg_Vecs& paVecs )
{
	//	assert only 1 input vector
	bassert ( paVecs.size() == 1 );

	//	assert sizes of stddev, output, input
	bassert ( m_stddev.size() == outVec.size() );
	bassert ( outVec.size() == paVecs.at(0).size() );
	
	double prob = 1.0;

	for ( unsigned int i = 0; i < m_stddev.size(); i++ )
	{
		if ( m_stddev.at(i) == 0 )
		{
			prob *= ( Is_equal(outVec.at(i), paVecs.at(0).at(i) - m_bias.at(i)) ) ? 1:0;
			//cout << "Branch 1: " << prob << endl;
		}
		else
		{
			//	Use helper function call to get the probability density at that point
			normal distr( paVecs.at(0).at(i) - m_bias.at(i), m_stddev.at(i) );
			//	Multiply all the probabilities together in one long chain
			prob *= pdf(distr, outVec.at(i));
			//cout << "Branch 2: " << pdf(distr, outVec.at(i)) << endl;
		}
	}
	
	bassert ( prob == prob );	// Make sure it is finite

	return prob;
}
// *************************************************************************
// *************************************************************************
void cpd_Gaussian::Unit_test ( void )
{
	//	Set up stddev and bias vectors
	cpd_Gaussian gaus( nfg_Vec(2,1) );
	//cpd_Gaussian gaus( nfg_Vec(2,1), nfg_Vec(2,5) );

	//	Set up parents vector list
	double mydbls[] = {10.5,1.0};
	nfg_Vecs paVecs(1,nfg_Vec(mydbls, mydbls+2)) ;
	cout << "Parent Vecs: ";
	Print_vec(paVecs.at(0));

	//	Test sample_node()
	nfg_Vec outVec = gaus.Sample_node( paVecs );
	cout << "Sampled Vecs: ";
	Print_vec(outVec);

	//	Test Get_prob_prop()
	cout << "Prob: " << gaus.Get_probability_prop( outVec,  paVecs ) << endl;

	cout << "Unit test completed." << endl << endl;
}

