
// *************************************************************************
//	LibNFG: Net-form game library
//
//	Author:		Ritchie Lee
//
//  Copyright 2011 United States Government as represented by the Administrator 
//  of the National Aeronautics and Space Administration.  All Rights Reserved.  
//  Licensed under the Apache License, Version 2.0 (the "License"); you may not 
//  use this file except in compliance with the License.  You may obtain a copy 
//  of the License at http://www.apache.org/licenses/LICENSE-2.0
//  Unless required by applicable law or agreed to in writing, software 
//  distributed under the License is distributed on an "AS IS" BASIS, WITHOUT 
//  WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the 
//  License for the specific language governing permissions and limitations 
//  under the License.
//
//	Revision Log
//
//	Username	Date		Description
//	-------------------------------------------------------
//	rlee		20110512	Created
//
// *************************************************************************

#include "cpd_tablevectorindep.h"

using namespace nfg;
using namespace boost;
using namespace boost::math;

// *************************************************************************
// *************************************************************************
cpd_TableVectorIndep::cpd_TableVectorIndep( const nfg_Dims& space, const nfg_Mask& paHasDno )
{
	// Input should be of the form [[self], [pa0], [pa1], [pa2] ...]
	// Each component of the above should be organized according to dimension:
	// e.g. [self] = [ sizeof(dim0), sizeof(dim1), sizeof(dim2) ...]
	
	// The extra component in space is [self].
	bassert( space.size()-1 == paHasDno.size() ); 

	// Save inputs to data members
	m_space			= space;
	m_paHasDno		= paHasDno;

	// Prep the data members to the right size
	m_ndims.resize(space.size());
	m_elements.resize(space.size());
	m_strides.resize(space.size());
	
	// Loop through each of self, pa0, pa1, etc.
	for ( unsigned int d = 0; d < space.size(); d++ )
	{
		// Save number of vector components in each of self, pa0, pa1, etc.
		m_ndims.at(d) = space.at(d).size();
		
		// number of elements is the product of sizes in each dimensions
		m_elements.at(d) = 1;
		for ( unsigned int i = 0; i < m_ndims.at(d); i++ )
			m_elements.at(d) *= space.at(d).at(i);

		// Insert an extra column for dno if it is declared
		if( d > 0 && m_paHasDno.at(d-1) )
			m_elements.at(d)++;

		// Figure out the stride length in each dimension
		for ( unsigned int i = 0; i < m_ndims.at(d); i++ )
		{
			unsigned int stride = 1;
			for ( unsigned int j = i+1; j < m_ndims.at(d); j++ )
			{
				stride *= space.at(d).at(j);
			}
			m_strides.at(d).push_back(stride);
		}
	}

	// Table size implied by m_space
	m_paTotalSize = 1;
	for( unsigned int i = 1; i < m_elements.size(); i++ )
		m_paTotalSize *= m_elements.at(i);
}
// *************************************************************************
// *************************************************************************
cpd_TableVectorIndep::~cpd_TableVectorIndep ( void )
{
}
// *************************************************************************
// *************************************************************************
void cpd_TableVectorIndep::Add_output_component ( const nfg_Vec& tbl )
{	
	// Can't add more components than size specified in m_space
	bassert( m_tables.size() < m_space.at(0).size() );

	// Make sure the table size is consistent with what was declared in m_space
	bassert( m_space.at(0).at(m_tables.size())*m_paTotalSize == tbl.size());
	
	// The size of the table is essentially the number of elements in each dimension
	nfg_Dim dims(m_elements);

	// except for outVec, which is the current size
	dims.at(0) = m_space.at(0).at( m_tables.size() );

	// Declare a new basic table for this dimension and add it to the list
	m_tables.push_back( cpd_TableBasic( dims, tbl ) );
}
// *************************************************************************
// *************************************************************************
unsigned int cpd_TableVectorIndep::Get_1d_index( const unsigned int dim, const nfg_Dim& idv )
{
	// First, make sure the request is valid
	assert ( Check_bounds(dim,idv) );

	unsigned int idx = 0;

	// empty() flag means dno, is stored as the last element
	if ( idv.empty() )
	{
		idx = m_elements.at(dim)-1;
	}
	else
	{
		// Convert tuple to linear indexing by multiplying by stride lengths
		for ( unsigned int i = 0; i < m_ndims.at(dim); i++ )
		{
			idx += idv.at(i)*m_strides.at(dim).at(i);
		}
	}

	return idx;
}
// *************************************************************************
// *************************************************************************
bool cpd_TableVectorIndep::Check_bounds( const unsigned int dim, const nfg_Dim& idv )
{
	bool isValid = true;

	// if the index to lookup is empty (i.e., flag for dno) then check to
	// see if dno is enabled for that dimension
	if ( idv.empty() )
	{
		isValid = m_paHasDno.at(dim-1);
	}
	// if idv is not dno, then do the size check
	else
	{
		// In order to be within bounds, it needs to be between 0 and size
		// in each dimension
		for ( unsigned int i = 0; i < m_ndims.at(dim); i++ )
			isValid = isValid && ( idv.at(i) < m_space.at(dim).at(i) ) && ( idv.at(i) >= 0 );
	}

	return isValid;
}
// *************************************************************************
// *************************************************************************
nfg_Vecs cpd_TableVectorIndep::Convert_vecs_to_basic ( const nfg_Vecs& vecs )
{
	// Make a new vector list to contain the basic representation of vecs
	nfg_Vecs vecsBasic( vecs.size() );

	// Convert each vec individually
	for( unsigned int i = 0; i < vecs.size(); i++ )
	{
		vecsBasic.at(i).push_back ( Get_1d_index( i+1, Cast_to_uint_vec(vecs.at(i))) );
	}

	return vecsBasic;
}
// *************************************************************************
// *************************************************************************
nfg_Vec	cpd_TableVectorIndep::Sample_node ( const nfg_Vecs& paVecs )
{
	// Make sure number of output components registered is same as declared
	bassert( m_tables.size() == m_space.at(0).size() );
	
	// Convert the parent vectors to basic representation
	nfg_Vecs paVecsBasic = Convert_vecs_to_basic( paVecs );

	// Sample each component of the output vector separately
	nfg_Vec outVec;
	for( unsigned int d = 0; d < m_ndims.at(0); d++ )
	{
		nfg_Vec basicVec = m_tables.at(d).Sample_node( paVecsBasic );
		outVec.push_back( basicVec.at(0) );
	}

	return outVec;
}
// *************************************************************************
// *************************************************************************
double cpd_TableVectorIndep::Get_probability ( const nfg_Vec& outVec, const nfg_Vecs& paVecs )
{	
	// Make sure number of output components registered is same as declared
	bassert( m_tables.size() == m_space.at(0).size() );
	
	// Convert the parent vectors to basic representation
	nfg_Vecs paVecsBasic = Convert_vecs_to_basic( paVecs );

	// Determine the probability of each component and multiply them together
	double prob = 1.0;
	for( unsigned int d = 0; d < m_ndims.at(0); d++ )
	{
		nfg_Vec outVecBasic(1, outVec.at(d));
		prob *= m_tables.at(d).Get_probability( outVecBasic, paVecsBasic );
	}

	return prob;
}
// *************************************************************************
// *************************************************************************
void cpd_TableVectorIndep::Unit_test ( void )
{
	nfg_Dims dims(3);
	dims.at(0) = nfg_Dim(2,3); // self: nDims 2, size of space 3 in each dim
	dims.at(1) = nfg_Dim(2,2); // pa0: nDims 2, size of space 2 in each dim
	dims.at(2) = nfg_Dim(2,2); // pa1: nDims 2, size of space 2 in each dim

	// Define CPDs
	// Col 1: [pa0(0,0),pa1(0,0)] 
	// Col 2: [pa0(0,0),pa1(0,1)]
	// Col 3: [pa0(0,0),pa1(1,0)]
	// Col 4: [pa0(0,0),pa1(1,1)]
	// Col 5: [pa0(0,0),pa1(dno)]
	// Col 6: [pa0(0,1),pa1(0,0)] 
	// Col 7: [pa0(0,1),pa1(0,1)]
	// Col 8: [pa0(0,1),pa1(1,0)]
	// Col 9: [pa0(0,1),pa1(1,1)]
	// Col 10: [pa0(0,1),pa1(dno)]
	// Col 11: [pa0(1,0),pa1(0,0)] 
	// Col 12: [pa0(1,0),pa1(0,1)]
	// Col 13: [pa0(1,0),pa1(1,0)]
	// Col 14: [pa0(1,0),pa1(1,1)]
	// Col 15: [pa0(1,0),pa1(dno)]
	// Col 16: [pa0(1,1),pa1(0,0)] 
	// Col 17: [pa0(1,1),pa1(0,1)]
	// Col 18: [pa0(1,1),pa1(1,0)]
	// Col 19: [pa0(1,1),pa1(1,1)]
	// Col 20: [pa0(1,1),pa1(dno)]
	// Column number			001, 002, 003, 004, 005, 006, 007, 008, 009, 010, 011, 012, 013, 014, 015, 016, 017, 018, 019, 020
	double		dim1Probs[]	= { 0.3, 0.2, 0.1, 0.7, 0.3, 0.7, 0.1, 0.5, 0.3, 0.5, 0.5, 0.5, 0.3, 0.7, 0.3, 0.5, 0.3, 0.4, 0.5, 0.6,
								0.4, 0.4, 0.2, 0.2, 0.3, 0.1, 0.1, 0.4, 0.4, 0.3, 0.2, 0.3, 0.3, 0.1, 0.1, 0.4, 0.2, 0.3, 0.2, 0.2,
								0.3, 0.4, 0.7, 0.1, 0.4, 0.2, 0.8, 0.1, 0.3, 0.2, 0.3, 0.2, 0.4, 0.2, 0.6, 0.1, 0.3, 0.4, 0.3, 0.2 };
	double		dim2Probs[]	= { 0.2, 0.8, 0.3, 0.2, 0.2, 0.8, 0.3, 0.2, 0.2, 0.8, 0.3, 0.2, 0.2, 0.8, 0.3, 0.2, 0.5, 0.3, 0.4, 0.5,
								0.2, 0.1, 0.4, 0.5, 0.5, 0.1, 0.1, 0.2, 0.2, 0.1, 0.4, 0.5, 0.5, 0.1, 0.1, 0.2, 0.4, 0.4, 0.2, 0.3,
								0.6, 0.1, 0.3, 0.3, 0.3, 0.1, 0.6, 0.6, 0.6, 0.1, 0.3, 0.3, 0.3, 0.1, 0.6, 0.6, 0.1, 0.3, 0.4, 0.2 };

	// Specify which parent has dno
	nfg_Mask paHasDno;
	paHasDno.push_back(false); paHasDno.push_back(true);

	// Construct object
	cpd_TableVectorIndep tvi(dims, paHasDno );
	// one for each output dimension
	tvi.Add_output_component( nfg_Vec(dim1Probs,dim1Probs+sizeof(dim1Probs)/sizeof(double)) );
	tvi.Add_output_component( nfg_Vec(dim2Probs,dim2Probs+sizeof(dim2Probs)/sizeof(double)) );

	// Set up parent vectors
	nfg_Vec paVec0;
	paVec0.push_back(1); paVec0.push_back(0);
	
	nfg_Vec paVec1;
	//paVec1.push_back(0); paVec1.push_back(0);
	paVec1.clear();

	nfg_Vecs paVecs(2);
	paVecs.at(0) = paVec0;
	paVecs.at(1) = paVec1;

	//	Test Sample_node()
	int N = 2000;
	nfg_Dim hist1(3,0);
	nfg_Dim hist2(3,0);
	nfg_Vec outVec;
	// Sample N times and collect satistics.  Make sure the fraction matches the
	// respective probabilities in the CPD table
	for ( int i = 0; i < N; i++ )
	{
		outVec = tvi.Sample_node ( paVecs );
		++hist1.at( (int)outVec.at(0) );
		++hist2.at( (int)outVec.at(1) );
	}
	cout << "Statistics:" << endl;
	cout << "Dim1..." << endl;
	cout << "0:" << (double)hist1.at(0) / N << endl;
	cout << "1:" << (double)hist1.at(1) / N << endl;
	cout << "2:" << (double)hist1.at(2) / N << endl;
	cout << "Dim2..." << endl;
	cout << "0:" << (double)hist2.at(0) / N << endl;
	cout << "1:" << (double)hist2.at(1) / N << endl;
	cout << "2:" << (double)hist2.at(2) / N << endl << endl;

	// Test Get_probability()
	// Manually (and arbitrarily) choose some outVec and paVecs to test
	outVec.resize(2); paVecs.at(0).resize(2); paVecs.at(1).resize(2);
	outVec.at(0) = 1; outVec.at(1) = 0; 
	paVecs.at(0).at(0) = 0; paVecs.at(0).at(1) = 1;
	paVecs.at(1).at(0) = 0; paVecs.at(1).at(1) = 1;
	cout << "P([1 0] | [0,1],[0,1]): " << tvi.Get_probability(outVec,paVecs) << endl; // P=0.1*0.3
	outVec.resize(2); paVecs.at(0).resize(2); paVecs.at(1).resize(2);
	outVec.at(0) = 2; outVec.at(1) = 0; 
	paVecs.at(0).at(0) = 1; paVecs.at(0).at(1) = 0;
	paVecs.at(1).at(0) = 0; paVecs.at(1).at(1) = 0;
	cout << "P([2,0] | [1,0],[0,0]): " << tvi.Get_probability(outVec,paVecs) << endl; // P=0.3*0.3
	outVec.resize(2); paVecs.at(0).resize(2); paVecs.at(1).resize(2);
	outVec.at(0) = 0; outVec.at(1) = 2; 
	paVecs.at(0).at(0) = 0; paVecs.at(0).at(1) = 0;
	paVecs.at(1).at(0) = 1; paVecs.at(1).at(1) = 1;
	cout << "P([0,2] | [0,0],[1,1]): " << tvi.Get_probability(outVec,paVecs) << endl; // P=0.7*0.3
	outVec.resize(2); paVecs.at(0).resize(2); paVecs.at(1).resize(2);
	outVec.at(0) = 0; outVec.at(1) = 1; 
	paVecs.at(0).at(0) = 1; paVecs.at(0).at(1) = 1;
	paVecs.at(1).clear();
	cout << "P([0,1] | [1,1],[dno]): " << tvi.Get_probability(outVec,paVecs) << endl;	// P=0.6*0.3
	
	cout << "Unit test completed." << endl << endl;

}

