
// *************************************************************************
//	LibNFG: Net-form game library
//
//	Author:		Ritchie Lee
//
//  Copyright 2011 United States Government as represented by the Administrator 
//  of the National Aeronautics and Space Administration.  All Rights Reserved.  
//  Licensed under the Apache License, Version 2.0 (the "License"); you may not 
//  use this file except in compliance with the License.  You may obtain a copy 
//  of the License at http://www.apache.org/licenses/LICENSE-2.0
//  Unless required by applicable law or agreed to in writing, software 
//  distributed under the License is distributed on an "AS IS" BASIS, WITHOUT 
//  WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the 
//  License for the specific language governing permissions and limitations 
//  under the License.
//
//	Revision Log
//
//	Username	Date		Description
//	-------------------------------------------------------
//	rlee		20101025	Created
//
// *************************************************************************

#include "cpd_uniform.h"

#include <iostream>

using namespace nfg;

// *************************************************************************
// *************************************************************************
cpd_Uniform::cpd_Uniform( const vector<Range>& range )
{
	//	Right now supports N-D continuous, not based on inputs
	
	//	Check in each dimension that min is smaller than max
	foreach ( Range ri, range )
		bassert ( ri.m_min < ri.m_max );

	//	Store the vector to member
	m_range = range;
}
// *************************************************************************
// *************************************************************************
cpd_Uniform::cpd_Uniform( const nfg_Vec& rangeMinVec, const nfg_Vec& rangeMaxVec )
{
	//	Right now supports N-D continuous, not based on inputs

	//	Check that the vector sizes match
	bassert ( rangeMinVec.size() == rangeMaxVec.size() );
	//	Check that they are not empty
	bassert ( !rangeMinVec.empty() );

	//	
	for ( unsigned int i = 0; i < rangeMinVec.size(); i++ )
	{
		//	Check in each dimension that min is smaller than max
		bassert ( rangeMinVec.at(i) < rangeMaxVec.at(i) );

		//	Recast the input format to data member format
		//	Create a range pair and store it to member
		Range rg(rangeMinVec.at(i), rangeMaxVec.at(i));
		m_range.push_back(rg);
	}

}
// *************************************************************************
// *************************************************************************
cpd_Uniform::~cpd_Uniform ( void )
{
}
// *************************************************************************
// *************************************************************************
nfg_Vec	cpd_Uniform::Sample_node ( const nfg_Vecs& paVecs )
{
	//	No support for input vecs. Priors only.
	//	Check that parents are empty
	bassert ( paVecs.empty() );

	nfg_Vec sample;		// to be output

	//	For each dimension, call helper function to generate a uniform random
	//	number (double) in the range of min/max.
	foreach ( Range ri, m_range )
		sample.push_back( Rand_uniform_range( ri.m_min, ri.m_max ) );

	return sample;
}
// *************************************************************************
// *************************************************************************
double cpd_Uniform::Get_probability_prop ( const nfg_Vec& outVec, const nfg_Vecs& paVecs)
{
	//	Used as a prior only
	bassert ( paVecs.empty() );

	bassert ( outVec.size() == m_range.size() );

	//	Check if inside area defined by m_range
	bool isInArea = true;
	for ( unsigned int i = 0; i < outVec.size(); i++ )
		isInArea = isInArea && Is_in_range ( outVec.at(i), m_range.at(i).m_min, m_range.at(i).m_max );

	//	If it is inside the bounds, figure out the area of the box.
	if ( isInArea )
	{
		double area = 1.0;

		foreach ( Range ri, m_range )
			area *= ri.m_max - ri.m_min;
		
		return 1 / area;
	}
	//	if not in bounds, probability of getting that number is 0
	else
		return 0.0;

}
// *************************************************************************
// *************************************************************************
void cpd_Uniform::Unit_test ( void )
{
	//	Set up cpd
	cpd_Uniform uni( vector<Range>(5,Range(0,2)) );
	nfg_Vecs paVecs;

	//	Test sample_node()
	nfg_Vec outVec = uni.Sample_node( paVecs );
	Print_vec(outVec);

	//	Test get_prob()
	cout << "Prob:" << uni.Get_probability_prop(outVec, paVecs) << endl;

	cout << "Unit test completed." << endl << endl;
}

