
// *************************************************************************
//	LibNFG: Net-form game library
//
//	Author:		Ritchie Lee
//
//  Copyright 2011 United States Government as represented by the Administrator 
//  of the National Aeronautics and Space Administration.  All Rights Reserved.  
//  Licensed under the Apache License, Version 2.0 (the "License"); you may not 
//  use this file except in compliance with the License.  You may obtain a copy 
//  of the License at http://www.apache.org/licenses/LICENSE-2.0
//  Unless required by applicable law or agreed to in writing, software 
//  distributed under the License is distributed on an "AS IS" BASIS, WITHOUT 
//  WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the 
//  License for the specific language governing permissions and limitations 
//  under the License.
//
//	Revision Log
//
//	Username	Date		Description
//	-------------------------------------------------------
//	rlee		20101025	Created
//	rlee		20110602	Added Convert_Dim_to_String and Unit_subtest1
//
// *************************************************************************

#include "helper.h"

#include <boost/lambda/lambda.hpp>
#include <boost/random.hpp>
#include <boost/math/distributions/normal.hpp>
#include <boost/math/distributions/poisson.hpp>
#include <boost/math/distributions/bernoulli.hpp>
#include <boost/math/special_functions/round.hpp>

using namespace std;
using namespace boost;
using namespace boost::lambda;

namespace nfg
{
	// *************************************************************************
	// *************************************************************************
	nfg_Dim Vectorize ( const int val1, const int val2 )
	{
		//	Populate vector
		nfg_Dim v_int;
		v_int.push_back(val1);
		v_int.push_back(val2);

		return v_int;
	}
	// *************************************************************************
	// *************************************************************************
	nfg_Dim Vectorize ( const int val1, const int val2, const int val3 )
	{
		//	Populate vector
		nfg_Dim v_int;
		v_int.push_back(val1);
		v_int.push_back(val2);
		v_int.push_back(val3);

		return v_int;
	}
	// *************************************************************************
	// *************************************************************************
	nfg_Dim Vectorize_discrete_range ( const int first, const int last )
	{
		// Check for valid input
		bassert( first == first );
		bassert( last == last );
		bassert( first <= last );

		//	Populate vector
		nfg_Dim v_int;

		for( int i = first; i <= last; i++ )
		{
			v_int.push_back(i);
		}

		return v_int;
	}
	// *************************************************************************
	// *************************************************************************
	double Rand_uniform_01 ( void )
	{
		//	set up the random seed
		static mt19937 rng(rand());
		//	call the boost variate generator
		variate_generator<mt19937&,uniform_01<double,double> > gen( rng, uniform_01<double,double>() );
		return gen();
	}
	// *************************************************************************
	// *************************************************************************
	double Rand_uniform_range ( const double min, const double max )
	{
		//	set up the random seed
		static mt19937 rng(rand());
		//	call the boost variate generator
		variate_generator<mt19937&,uniform_real<double> > gen ( rng, uniform_real<double>(min, max) );
		return gen();
	}
	// *************************************************************************
	// *************************************************************************
	double Rand_gaussian ( const double mean, const double stddev )
	{
		//	set up the random seed
		static mt19937 rng(rand());
		//	call the boost variate generator
		variate_generator<mt19937&,boost::normal_distribution<double> > gen( rng, normal_distribution<double>(mean, stddev) );
		return gen();
	}
	// *************************************************************************
	// *************************************************************************
	double Rand_poisson ( const double meanRate )
	{
		//	set up the random seed
		static mt19937 rng(rand());
		//	call the boost variate generator
		variate_generator<mt19937&,poisson_distribution<> > gen( rng, poisson_distribution<>(meanRate) );
		return gen();
	}
	// *************************************************************************
	// *************************************************************************
	double Rand_bernoulli ( const double probSuccess )
	{
		//	set up the random seed
		static mt19937 rng(rand());
		//	call the boost variate generator
		variate_generator<mt19937&,bernoulli_distribution<> > gen( rng, bernoulli_distribution<>(probSuccess) );
		return gen();
	}
	// *************************************************************************
	// *************************************************************************
	int Rand_int ( const int min, const int max )
	{
		//	set up the random seed
		static mt19937 rng(rand());
		//	call the boost variate generator
		variate_generator<mt19937&,boost::uniform_int<> > gen( rng, boost::uniform_int<>(min, max) );
		return gen();
	}
	// *************************************************************************
	// *************************************************************************
	double Indicator_iround ( const int nodeId, const unsigned int value, const nfg_Inst& instVecs )
	{
		int rval = Cast_to_uint( instVecs.at(nodeId).at(0) );
		return rval == value ? 1.0 : 0;
	}
	// *************************************************************************
	// *************************************************************************
	unsigned int Cast_to_uint ( const double val )
	{
		bassert ( abs( math::iround(val) - val ) < DISCRETE_ROUNDING_TOL );
		return  math::iround(val);
	}
	// *************************************************************************
	// *************************************************************************
	nfg_Dim Cast_to_uint_vec ( const nfg_Vec& dvec )
	{
		nfg_Dim ivec;
		foreach( double d, dvec )
			ivec.push_back( Cast_to_uint(d) );

		return  ivec;
	}
	// *************************************************************************
	// *************************************************************************
	bool Is_equal ( const double x, const double y )
	{
		return ( abs(x - y) < LIB_ROUNDING_TOL );
	}
	// *************************************************************************
	// *************************************************************************
	bool Is_equal ( const nfg_Vec& vec1, const nfg_Vec& vec2 )
	{
		// If sizes are not equal, then they are automatically not equal
		if( vec1.size() != vec2.size() )
			return false;
		
		// Loop through all values and make sure all values are equal pairwise
		bool overall = true;
		for( unsigned int i = 0; i < vec1.size(); i++ )
			overall &= Is_equal( vec1.at(i), vec2.at(i) );

		return overall;
	}
	// *************************************************************************
	// *************************************************************************
	bool Is_discrete_equal ( const double x, const double y )
	{
		return ( abs(x - y) < DISCRETE_ROUNDING_TOL );
	}
	// *************************************************************************
	// *************************************************************************
	bool Is_in_range ( const double val, const double rangeMinVec, const double rangeMaxVec )
	{
		return val >= rangeMinVec && val <= rangeMaxVec;
	}
	// *************************************************************************
	// *************************************************************************
	void Print_vec( const nfg_Vec& vec )
	{
		//	Output vector to screen
		cout << "<";
		for_each ( vec.begin(), vec.end(), cout << " " << _1 << ", " );
		cout << ">" << endl;
	}
	// *************************************************************************
	// *************************************************************************
	void Print_veclist( nfg_Vecs veclist )
	{
		//	Repeatedly call print_vec on each vec
		foreach( nfg_Vec v, veclist )
			Print_vec( v );

		cout << endl;
	}
	// *************************************************************************
	// *************************************************************************
	int Argmax ( const nfg_Vec& vec )
	{
		double max = vec.at(0);		// Initialize to the first value
		int max_i = 0;				// Remember the index of the max value	
		int i = 0;					// Track current index
		foreach( double d, vec )
		{
			//	Check for valid numbers
			bassert( d==d );

			if ( d > max )
			{
				max = d;
				max_i = i;
			}

			i++;
		}

		return max_i;	// return index at which the maximum occurred
	}
	// *************************************************************************
	// *************************************************************************
	nfg_Vec Exponentiate ( const nfg_Vec& vec, const double lambda )
	{
		nfg_Vec outVec = vec;
	
		//	Perform element-wise exponentiation in the form exp(lambda*x)
		foreach( double& d, outVec )
		{
			d = exp(lambda*d);
		}

		return outVec;
	}
	// *************************************************************************
	// *************************************************************************
	nfg_Vec Normalize ( const nfg_Vec& vec )
	{
		nfg_Vec outVec = vec;	// Start with a copy of the input vector
		double normConst = 0;	// Normalization constant

		//	Determine normalization constant
		foreach( double d, outVec )
		{
			normConst += d;		// Sum each component
		}

		bassert ( normConst > 0 );		// Make sure the norm constant is valid

		//	Perform actual normalization
		foreach( double& d, outVec )
		{
			d /= normConst;			// Divide each element by the normalization const
		}

		return outVec;
	}
	// *************************************************************************
	// *************************************************************************
	nfg_Vec Convert_pdf_to_cdf ( const nfg_Vec& pdf )
	{
		nfg_Vec cdf(pdf);	// Start with a copy of the pdf
		double sum = 0.0;	// Keep track of a running sum

		foreach ( double& prob, cdf )
		{
			sum += prob;	// Keep a running sum
			prob = sum;		// Assign the current sum to the element
		}

		//	Complain if the probabilities don't add up to 1 (within a tolerance)
		bassert( Is_discrete_equal(sum, 1.0) );

		return cdf;
	}
	// *************************************************************************
	// *************************************************************************
	int Weighted_sample ( const nfg_Vec& cum_distr )
	{
		bassert ( !cum_distr.empty() );		// Check that cdf is not empty
		// Check that it is a valid cdf, i.e., last element should be 1.0
		bassert ( Is_discrete_equal(cum_distr.back(),1.0) );
		
		// Generate a random number between 0 and 1...
		double rn = Rand_uniform_01();

		// And scroll through until the random number exceeds the value or reach
		// the end.
		unsigned int id;
		for ( id = 0; id < cum_distr.size() && rn > cum_distr.at(id); id++ );
		
		// Return where it stopped
		return id;
	}
	// *************************************************************************
	// *************************************************************************
	void Sort_and_remove_duplicates ( nfg_Vec& vec )
	{
		// Call std algorithm functions to sort and make unique
		sort ( vec.begin(), vec.end() );
		nfg_Vec::iterator it = unique ( vec.begin(), vec.end() );
		vec.resize( it - vec.begin() );
	}
	// *************************************************************************
	// *************************************************************************
	void Sort_and_remove_duplicates ( nfg_Vecs& veclist )
	{
		// Call std algorithm functions to sort and make unique
		sort ( veclist.begin(), veclist.end() );
		nfg_Vecs::iterator it = unique ( veclist.begin(), veclist.end() );
		veclist.resize( it - veclist.begin() );
	}
	// *************************************************************************
	// *************************************************************************
	nfg_Space1D Construct_discrete_space1d ( const nfg_Dim& vec )
	{
		// Make sure the vector isn't empty
		bassert( !vec.empty() );

		nfg_Space1D subspace;

		foreach( int z, vec )
		{
			subspace.push_back( pair<double,double>(z,z) );
		}

		return subspace;
	}
	// *************************************************************************
	// *************************************************************************
	nfg_Inst Convert_OptInst_to_Inst ( const nfg_OptInst& optInst )
	{
		// Container for complete weighted instantiation
		nfg_Inst inst;

		// Make sure it is a complete instantiation
		foreach( optional<nfg_Vec> optVec, optInst )
		{
			bassert ( optVec );				// make sure that vector is defined
			inst.push_back( *optVec );		// copy it over to the output one
		}

		return inst;
	}
	// *************************************************************************
	// *************************************************************************
	string Convert_Dim_to_String ( const nfg_Dim& vec )
	{
		string out;

		foreach( unsigned int i, vec )
			out.append( lexical_cast<string>(i) );

		return out;
	}
	// *************************************************************************
	// *************************************************************************
	void Unit_subtest1 ( void )
	{
		//	Test Rand_int

		cout << Rand_int(1,5) << endl;	
		cout << Rand_int(1,5) << endl;	
		cout << Rand_int(1,5) << endl;
		cout << Rand_int(1,5) << endl;	
		cout << Rand_int(1,5) << endl;	
		cout << Rand_int(1,5) << endl;
	}
	// *************************************************************************
	// *************************************************************************
	void Unit_subtest2 ( void )
	{
		//	Test Sort_and_remove_duplicates

		// Load up some vectors
		nfg_Vec vec1; vec1.push_back(1.0); vec1.push_back(2.0);
		nfg_Vec vec2; vec2.push_back(2.0); vec2.push_back(1.0);
		nfg_Vec vec3; vec3.push_back(2.0); vec3.push_back(2.0);

		nfg_Vecs vecList;
		vecList.push_back( vec1 );
		vecList.push_back( vec1 );
		vecList.push_back( vec2 );
		vecList.push_back( vec3 );
		vecList.push_back( vec2 );
		vecList.push_back( vec1 );
		
		// Will sort by first element, then second, etc.
		// Correct order is vec1, vec2, vec3.
		Sort_and_remove_duplicates( vecList );

		return;
	}
	// *************************************************************************
	// *************************************************************************
	void Unit_subtest3 ( void )
	{
		// Load up a double vector
		nfg_Vec vec;
		vec.push_back(1);
		vec.push_back(2);
		vec.push_back(3);
		vec.push_back(2);
		vec.push_back(1);
		
		// convert it to int vector
		// then convert it to string
		// and output it to screen
		cout << Convert_Dim_to_String( Cast_to_uint_vec(vec) ) << endl;

		return;
	}
	// *************************************************************************
	// *************************************************************************
	void Unit_subtest4 ( void )
	{
		// Load up test vectors
		nfg_Vec v1;
		v1.push_back(1.0);
		v1.push_back(3.0);
		v1.push_back(5.0);

		nfg_Vec v2;
		v2.push_back(1.0);
		v2.push_back(3.0);
		v2.push_back(2.0);

		// Check that these are equal
		cout << ( Is_equal(v1,v1) ? "true":"false" ) << endl;

		// Check that these are not equal
		cout << ( Is_equal(v1,v2) ? "true":"false" ) << endl;

		return;
	}
}	


