
// *************************************************************************
//	LibNFG: Net-form game library
//
//	Author:		Ritchie Lee
//
//  Copyright 2011 United States Government as represented by the Administrator 
//  of the National Aeronautics and Space Administration.  All Rights Reserved.  
//  Licensed under the Apache License, Version 2.0 (the "License"); you may not 
//  use this file except in compliance with the License.  You may obtain a copy 
//  of the License at http://www.apache.org/licenses/LICENSE-2.0
//  Unless required by applicable law or agreed to in writing, software 
//  distributed under the License is distributed on an "AS IS" BASIS, WITHOUT 
//  WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the 
//  License for the specific language governing permissions and limitations 
//  under the License.
//
//	Revision Log
//
//	Username	Date		Description
//	-------------------------------------------------------
//	rlee		20101025	Created
//	rlee		20110602	Added Convert_Dim_to_String and Unit_subtest1
//
// *************************************************************************

#ifndef INCLUDED_HELPER_H
#define INCLUDED_HELPER_H

#include "types.h"

namespace nfg
{
	//	Used for comparing two doubles to be considered equal.  
	const double	LIB_ROUNDING_TOL = 1e-50;

	//	Used in functions that require rounding from double to an int.
	//	Dealing with discrete variables stored as doubles.
	const double	DISCRETE_ROUNDING_TOL = 1e-5;

	/**
	 * Vectorize: Takes 2 ints and gives you a vector of them
	 *	val1 - first int
	 *	val2 - second int
	 *	returns: an int vector with val1 at(0) and val2 at(1)
	 */
	nfg_Dim		Vectorize ( const int val1, const int val2 );
	/**
	 * Vectorize: Takes 3 ints and gives you a vector of them
	 *	val1 - first int
	 *	val2 - second int
	 *	val3 - third int
	 *	returns: an int vector with val1 at(0), val2 at(1), and val3 at(2)
	 */
	nfg_Dim		Vectorize ( const int val1, const int val2, const int val3 );
	/**
	 * Vectorize_discrete_range: Create a vector of ints between first and last inclusive
	 *	first - first integer to include
	 *	last - last integer to include
	 *	returns: an int vector containing all integers between first and last inclusive
	 */
	nfg_Dim			Vectorize_discrete_range ( const int first, const int last );
	/**
	 * Rand_uniform_01: Generates a uniformly random double between 0 and 1
	 *	returns: a uniformly random double between 0 and 1
	 */
	double			Rand_uniform_01 ( void );
	/**
	 * Rand_uniform_range: Generates a uniformly random double between min and max
	 *	min - lowerbound on random number
	 *	max - upperbound on random number
	 *	returns: a uniformly random double between min and max
	 */
	double			Rand_uniform_range ( const double min, const double max );
	/**
	 * Rand_gaussian: Generates a Gaussian random double centered at mean with variance 'variance'
	 *	mean - mean value of Gaussian distribution
	 *	stddev - standard deviation (sigma) value of Gaussian distribution
	 *	returns: a Gaussian random double centered at mean with variance 'variance'
	 */
	double			Rand_gaussian ( const double mean, const double stddev );
	/**
	 * Rand_poisson: Generates a Poisson random double with mean 'mean'
	 *	meanRate - aka lambda. mean rate of Poisson distribution
	 *	returns: a Poisson random double with mean 'mean'
	 */
	double			Rand_poisson ( const double meanRate );
	/**
	 * Rand_bernoulli: Generates a Bernoulli random variable
	 *	probSuccess - aka p.  Probability of success.
	 *	returns: a Bernoulli random double with probability of success probSuccess
	 */
	double			Rand_bernoulli ( const double probSuccess );
	/**
	 * Rand_int: Generates a uniformly random int between min and max
	 *	min - lowerbound on random number
	 *	max - upperbound on random number
	 *	returns: a uniformly random int between min and max
	 */	
	int				Rand_int ( const int min, const int max );
	/**
	 * Indicator_iround: Indicator function with rounding and node lookup into a vector for comparing values.
	 *	nodeId - index into inst to compare to value.
	 *	value - target value to compare to
	 *	instVecs - list of vectors into which will be indexed
	 *	returns: 1.0 if compare is true, 0.0 if compare is false.
	 */
	double			Indicator_iround ( const int nodeId, const unsigned int value, const nfg_Inst& instVecs );
	/**
	 * Cast_to_int: Round a double to nearest int.  Value must be within rounding tolerance of
	 *		int.  Otherwise will fail.
	 *	val - double to be casted
	 *	returns: casted integer
	 */
	unsigned int Cast_to_uint ( const double val );
	/**
	 * Cast_to_uint_vec: Converts a double vector to a uint vector.  Calls Cast_to_uint.
	 *	dvec - vector of doubles to be converted
	 *	returns: converted vector of unsigned int
	 */
	nfg_Dim		Cast_to_uint_vec ( const nfg_Vec& dvec );
	/**
	 * Is_equal: Test whether 2 doubles are within tolerance of each other
	 *		to be treated as equal using LIB_ROUNDING_TOLERANCE.
	 *	x - test value 1
	 *	y - test value 2
	 *	returns: true if x and y are approximately equal, false otherwise
	 */
	bool		Is_equal ( const double x, const double y );
	/**
	 * Is_equal: Test whether 2 vectors of doubles are within tolerance of each other
	 *		to be treated as equal using LIB_ROUNDING_TOLERANCE.
	 *	vec1 - test vector 1
	 *	vec2 - test vector 2
	 *	returns: true if vec1 and vec2 are approximately equal, false otherwise
	 */
	bool		Is_equal ( const nfg_Vec& vec1, const nfg_Vec& vec2 );
	/**
	 * Is_discrete_equal: Test whether 2 doubles are within tolerance of each other
	 *		to be treated as equal using DISCRETE_ROUNDING_TOLERANCE.  Used in
	 *		functions for rounding doubles to ints (and making sure it's reasonable.)
	 *	x - test value 1
	 *	y - test value 2
	 *	returns: true if x and y are approximately equal, false otherwise
	 */
	bool		Is_discrete_equal ( const double x, const double y );
	/**
	 * Is_in_range: Test whether a number is inside a range.
	 *	val - test value
	 *	rangeMinVec - range minimum
	 *	rangeMaxVec - range maximum
	 *	returns: true if val is between min and max, false otherwise
	 */
	bool			Is_in_range ( const double val, const double rangeMinVec, const double rangeMaxVec );
	/**
	 * Print_vec: Output a vector to console
	 *	vec: vector to be output
	 */
	void			Print_vec( const nfg_Vec& vec );
	/**
	 * Print_veclist: Output a vector list to console
	 *	veclist - list of vectors to be output
	 */
	void			Print_veclist( nfg_Vecs veclist );
	/**
	 * Argmax: Finds the index at which the vector has maximum value
	 *	vec - list of vectors to be output
	 *	returns: index at which the vector has maximum value
	 */
	int				Argmax ( const nfg_Vec& vec );
	/**
	 * Exponentiate: Multiplies a vector by lambda and exponentiates each component
	 *	vec: vector to be exponentiated
	 *	lambda: exponential parameter in the exponent
	 *	returns: exp(lambda.*vec)
	 */
	nfg_Vec			Exponentiate ( const nfg_Vec& vec, const double lambda = 1.0 );
	/**
	 * Normalize: Normalizes a vector to sum to 1.0
	 *	vec - vector to be normalized
	 *	returns: exp(lambda.*vec)
	 */
	nfg_Vec			Normalize ( const nfg_Vec& vec );
	/**
	 * Weighted_sample: Performs a discrete weighted sample according to a cdf.  
	 *	cum_distr - vector of cumulative distribution function
	 *	returns: int associated with the sampled position.
	 */
	int				Weighted_sample ( const nfg_Vec& cum_distr );
	/**
	 * Convert_pdf_to_cdf: Converts a probability distribution function to a cumulative mass 
	 *		function for discrete variables 
	 *	pdf - vector representing probability distribution function
	 *	returns: vector representing cdf
	 */
	nfg_Vec			Convert_pdf_to_cdf ( const nfg_Vec& pdf );
	/**
	 * Sort_and_remove_duplicates: Sorts and remove duplicates for a nfg_Vec object.
	 *		operates directly on the object
	 *	vec - vector (by reference object) of interest, will be modified
	 */
	void			Sort_and_remove_duplicates ( nfg_Vec& vec );
	/**
	 * Sort_and_remove_duplicates: Sorts and remove duplicates for a nfg_Vecs object.
	 *		operates directly on the object
	 *	veclist: vector list (by reference object) of interest, will be modified
	 */
	void			Sort_and_remove_duplicates ( nfg_Vecs& veclist );
	/**
	 * Construct_discrete_space1d: Shortcut to construct a space1d object (object that
	 *		that represents a 1d discrete space) using a vector of doubles
	 *	vec - vector list of valid discrete values
	 *	returns: space1d object that represents the valid discrete values in vec
	 */
	nfg_Space1D		Construct_discrete_space1d ( const nfg_Dim& vec );
	/**
	 * Convert_OptInst_to_Inst: Convert an nfg_OptInst object to nfg_Inst
	 *		making sure that all values are valid (but does not check for size).  
	 *		vector.empty() entries are "does not occur" flags
	 *	optInst - optional instantiation object to be converted
	 *	returns: a copy of optInst with the opt part removed.
	 */
	nfg_Inst		Convert_OptInst_to_Inst ( const nfg_OptInst& optInst );
	/**
	 * Convert_Dim_to_String: Convert an vector of ints to a single string
	 *	vec - vector of ints to be concatenated
	 *	returns: converted string.
	 */
	string			Convert_Dim_to_String ( const nfg_Dim& vec );
	//	Debug
	void			Unit_subtest1 ( void );	// Test Rand_int()
	void			Unit_subtest2 ( void );	// Test Sort_and_remove_duplicates()
	void			Unit_subtest3 ( void );	// Test Convert_Vec_to_String()
	void			Unit_subtest4 ( void );	// Test Is_equal() for nfg_Vecs
}


#endif

