
// *************************************************************************
//	LibNFG: Net-form game library
//
//	Author:		Ritchie Lee
//
//  Copyright 2011 United States Government as represented by the Administrator 
//  of the National Aeronautics and Space Administration.  All Rights Reserved.  
//  Licensed under the Apache License, Version 2.0 (the "License"); you may not 
//  use this file except in compliance with the License.  You may obtain a copy 
//  of the License at http://www.apache.org/licenses/LICENSE-2.0
//  Unless required by applicable law or agreed to in writing, software 
//  distributed under the License is distributed on an "AS IS" BASIS, WITHOUT 
//  WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the 
//  License for the specific language governing permissions and limitations 
//  under the License.
//
//	Revision Log
//
//	Username	Date		Description
//	-------------------------------------------------------
//	rlee		20120710	Created
//
// *************************************************************************

#ifndef INCLUDED_IBAYESNET_H
#define INCLUDED_IBAYESNET_H

#include "types.h"
#include "helper.h"
#include "bayesnet.h"
#include "iplayer.h"
#include "cpd_glue.h"

class IBayesNet
{
public:

	/**	Constructor */
	IBayesNet( const string ibnetName, 
				BayesNet& baseBNet, BayesNet& kernelBNet, 
				const nfg_Edges& base2kernelEdges,
				const nfg_Edges& kernel2kernelEdges,
				const unsigned int numSteps );

	/**	Destructor */
	virtual ~ IBayesNet();

	/// Variables

	///	Methods
	/**
	 * Get_ibnet_name: Get the name of the ibnet
	 *	returns: name as a string
	 */
	string	Get_ibnet_name ( void ) { return m_ibnetName; };
	/**
	 * Get_num_steps: Get the number steps / times to repeat the net
	 *	returns: number of steps  / times to repeat the net
	 */
	unsigned int	Get_num_steps ( void ) { return m_numSteps; };
	/**
	 * Get_base_bnet: Get base bayes net object
	 *	returns: base bayes net object
	 */
	BayesNet&	Get_base_bnet ( void ) { return m_baseBNet; };
	/**
	 * Get_kernel_bnet: Get base bayes net object
	 *	returns: base bayes net object
	 */
	BayesNet&	Get_kernel_bnet ( void ) { return m_kernelBNet; };
	/**
	 * Get_base2kernelEdges: Get edges that connect base to kernel bnet
	 *	returns: edges that connect base to kernel bnet
	 */
	nfg_Edges	Get_base2kernelEdges ( void ) { return m_base2kernelEdges; };
	/**
	 * Get_kernel2kernelEdges: Get edges that connect kernel bnets
	 *	returns: edges that connect kernel bnets
	 */
	nfg_Edges	Get_kernel2kernelEdges ( void ) { return m_kernel2kernelEdges; };
	/**
	 * Add_player: Add a player to the list 
	 *	player - player object to be added
	 */
	void	Add_iplayer ( const IPlayer& iplayer ) { m_iplayers.push_back(iplayer); };
	/**
	 * Get_player_list: Get the entire player list
	 *	returns: player list
	 */
	vector<IPlayer>	Get_player_list ( void ) { return m_iplayers; };
	/**
	 * Clear_player_list: Clear the player list
	 */
	void	Clear_player_list ( void ) { m_iplayers.clear(); };
	/**
	 * Get_num_players: Get the number of players in the list
	 *	returns: player object to be added
	 */
	unsigned int	Get_num_players ( void ) { return m_iplayers.size(); };
	/**
	 * Set_player_policyFcn: Set player behavior
	 *	playerId - player number in the list
	 *	policyFunc - function that maps paVecs to outVec
	 */
	void	Set_player_policy ( int playerId, nfg_SampleFunc fo_policy );
	/**
	 * Forward_sample: sample ibnet
	 *	returns: instantiation of ibnet
	 */
	nfg_Episode		Forward_sample ( void );


private:

	/// Variables
	string							m_ibnetName;			// Name of the IBayesNet
	BayesNet						m_baseBNet;				// Base bayes net
	BayesNet						m_kernelBNet;			// Kernel bayes net
	nfg_Edges						m_base2kernelEdges;		// List of edges connecting base bayes net to kernel
															// bayes net.  i.e., first slice to second slice
	nfg_Edges						m_kernel2kernelEdges;	// List of edges connecting kernel bayes net to 
															// kernel bayes net, i.e., slices after first
	unsigned int					m_numSteps;				// Number of times to repeat the net treating bnet
															// as a step
	vector<IPlayer>					m_iplayers;				// List of iplayers


	/// Methods

};

#endif

