
// *************************************************************************
//	LibNFG: Net-form game library
//
//	Author:		Ritchie Lee
//
//  Copyright 2011 United States Government as represented by the Administrator 
//  of the National Aeronautics and Space Administration.  All Rights Reserved.  
//  Licensed under the Apache License, Version 2.0 (the "License"); you may not 
//  use this file except in compliance with the License.  You may obtain a copy 
//  of the License at http://www.apache.org/licenses/LICENSE-2.0
//  Unless required by applicable law or agreed to in writing, software 
//  distributed under the License is distributed on an "AS IS" BASIS, WITHOUT 
//  WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the 
//  License for the specific language governing permissions and limitations 
//  under the License.
//
//	Revision Log
//
//	Username	Date		Description
//	-------------------------------------------------------
//	rlee		20110420	Created
//	rlee		20110602	Updated to current definition of branch nodes
//
// *************************************************************************

#include "if_gambit.h"
#include <string>
#include <iostream>
#include <fstream>
#include <boost/algorithm/string.hpp>
#include <boost/optional.hpp>

using namespace std;
using namespace nfg;

// *************************************************************************
// *************************************************************************
if_Gambit::if_Gambit( void )
{
}
// *************************************************************************
// *************************************************************************
if_Gambit::~if_Gambit ( void )
{
}
// *************************************************************************
// *************************************************************************
void if_Gambit::Export_to_efg_format( BayesNet& bnet, const string filename )
{
	// Check that all nodes are discrete
	foreach ( nfg_Vertex vi, vertices(bnet.Get_graph()) )
		bassert ( !bnet.Get_vertex_cts(vi) );

	// Declare filestream
	ofstream outfile;

	// Open file to write
	outfile.open( filename.c_str() );
		
	if( outfile.is_open() )
	{
		// Output prologue
		outfile << "EFG 2 R \"" << bnet.Get_bnet_name() << "\" { ";

		// A "player" in Gambit is a "decision node" in libnfg
		// make a map of all the decision nodes
		nfg_Vertices dNodeList;
		foreach ( nfg_Vertex vi, vertices(bnet.Get_graph()) )
		{
			if ( bnet.Get_vertex_type(vi) == DECISION )
			{
				// vector index is player number in gambit
				// data element is vertex id in libnfg
				dNodeList.push_back(vi);
				// output to gambit's player list
				outfile << "\"" << bnet.Get_vertex_name(vi) << "\" ";
			}
		}
		outfile << "}" << endl;	// Mark end of prologue

		// Output each node line by line using a recursive function
		// enumerating all possible values of the net
		unsigned int tNum = 0;	// for creating unique terminal numbers
		unsigned int cNum = 0;	// for creating unique chance infosets
		Expand_node ( outfile, bnet, 0, 
				nfg_OptInst(bnet.Get_num_vertices(), 0), // expand first node in vertex list
				tNum, cNum, dNodeList );				// will call others recursively


		// Close file
		outfile.close();
	}
	else
	{
		cout << "\n\nERROR! Could not open output file: " << filename << endl;
		bassert( false );
	}

	return;
}
// *************************************************************************
// *************************************************************************
void if_Gambit::Expand_node ( ofstream& outfile, BayesNet& bnet, const nfg_Vertex v, 
								const nfg_OptInst& prevInst, unsigned int& tNum,
								unsigned int& cNum,
								const nfg_Vertices& dNodes )
{
	//	Error: should not have counted this high
	bassert ( v <= bnet.Get_num_vertices() );

	if ( v == bnet.Get_num_vertices() )
	{
		// Terminal node reached, write terminal line to file
		Write_line_terminal ( outfile, bnet, v, prevInst, ++tNum, dNodes );
	}
	else
	{
		// make a copy of what we've got so far in the instantiation
		nfg_OptInst curInst = prevInst;

		// If not disabled...
		if ( !bnet.Get_vertex_dno(v) || !bnet.Get_dno_func(v)(curInst) )
		{
			// check parents to determine information sets
			// separate into infosets
			// I'm using something similar to Horner's rule to calculate the
			// infoset number so that it doesn't overlap
			unsigned int pNum = 0;	// represents infoset number
			unsigned int prevSize = 0;	// size of the previous vertex (space allocation)
			foreach ( nfg_Vertex vpi, inv_adjacent_vertices(v, bnet.Get_graph()) )
			{
				pNum = pNum * prevSize + Cast_to_uint((*prevInst.at(vpi)).at(0));
				prevSize = bnet.Get_vertex_size(vpi);
			}

			// Write a node line using chance or player format
			switch ( bnet.Get_vertex_type(v) )
			{
			case CHANCE:
				// Chance format
				Write_line_chance ( outfile, bnet, v, prevInst, ++cNum, dNodes );
				break;
			case DECISION:
				// Player format
				Write_line_player ( outfile, bnet, v, prevInst, ++pNum, dNodes );
				break;
			default:
				bassert(false); // others not supported
			}

			// Use an iterator to go through the space
			DiscreteSpaceIterator it( bnet.Get_vertex_space(v) );
			// Expand each possible state of this node
			while ( !it().empty() )
			{
				// Enumerate each possible state value and branch recursively
				curInst.at(v) = it();	// update current instance with candidate value
				Expand_node( outfile, bnet, v+1, curInst, tNum, cNum, dNodes );
				++it;		// increment the iterator
			}
		}
		else
		{
			// If node was disabled, don't do any of the above, simply 
			// set the vector as empty in the instantiation and expand the next node.
			// No line is written
			curInst.at(v) = nfg_Vec();	// flag as disabled/dno
			Expand_node( outfile, bnet, v+1, curInst, tNum, cNum, dNodes );
		}
	}
}
// *************************************************************************
// *************************************************************************
optional<int> if_Gambit::Find_value( const nfg_Vertices& vlist, const nfg_Vertex val )
{
	// object for the index to be found
	optional<int> index;

	// look through the list
	for ( unsigned int i = 0; i < vlist.size(); i++ )
	{
		// if a matching value is found
		if ( vlist.at(i) == val )
		{
			// remember the index
			index = i;
			break;
		}
	}

	// return index.  Is an optional object.  Will be uninitialized if not found.
	// *index to get value if (index == true).
	return index;
}
// *************************************************************************
// *************************************************************************
void if_Gambit::Write_line_chance( ofstream& outfile, BayesNet& bnet, const nfg_Vertex v,
								  const nfg_OptInst& prevInst, unsigned int& cNum,
								  const nfg_Vertices& dNodes )
{
	// Output a c for chance. Then "vertex name". infoset number, then "infoset name" (empty)
	// then { "name of action1" probability1 "name of action2" probability2 etc}
	outfile << "c \"" << bnet.Get_vertex_name(v) << "\" ";
	outfile << cNum; // infoset;
	outfile << " \"\" { ";
	
	// Use an iterator to go through the space
	DiscreteSpaceIterator it( bnet.Get_vertex_space(v) );
	// for each outcome (state of the node), get the probability 
	while ( !it().empty() )
	{
		outfile << "\"" << Convert_Dim_to_String( Cast_to_uint_vec(it()) )
			<< "\" " << bnet.Get_cpd(v)->Get_probability( it(), 
					Get_parent_values( bnet, prevInst, v ) ) << " ";
		++it;	// increment the iterator
	}

	// close bracket and indicate 0 incremental payoff
	outfile << "} 0" << endl;
}
// *************************************************************************
// *************************************************************************
void if_Gambit::Write_line_player( ofstream& outfile, BayesNet& bnet, const nfg_Vertex v,
								  const nfg_OptInst& prevInst, unsigned int& pNum,
								  const nfg_Vertices& dNodes )
{
	// Output a p for player. Then "vertex name". 
	outfile << "p \"" << bnet.Get_vertex_name(v) << "\" ";
	
	// player number, infoset number, then "infoset name" (empty)
	optional<int> p = Find_value( dNodes, v );
	bassert ( p );	// Error: there's no reason why it was not found
	
	outfile << *p+1 << " ";

	outfile << pNum;
	outfile << " \"\"";	// empty infoset label
	outfile << " { ";
	
	// Use an iterator to go through the space
	DiscreteSpaceIterator it( bnet.Get_vertex_space(v) );
	// for each outcome (state of the node), output the name of the move
	while ( !it().empty() )
	{
		outfile << "\"" << Convert_Dim_to_String( Cast_to_uint_vec( it() ) ) << "\" ";
		++it;	// Increment the iterator
	}

	// close bracket and indicate 0 incremental payoff
	outfile << "} 0" << endl;
}
// *************************************************************************
// *************************************************************************
void if_Gambit::Write_line_terminal( ofstream& outfile, BayesNet& bnet, const nfg_Vertex v,
								  const nfg_OptInst& prevInst, unsigned int& tNum,
								  const nfg_Vertices& dNodes )
{
	// Check size of inst
	bassert( prevInst.size() == bnet.Get_num_vertices() );
	
	// character t for terminal node, terminal number, "terminal name"
	outfile << "t \"\" " << tNum << " \"Outcome " << tNum << "\" { ";

	// for each decision node (or player in gambit) output the payoff
	foreach ( nfg_Vertex dNode, dNodes )
	{
		// Get utility from decision node cpds, give it the full instantiation
		double util = bnet.Get_cpd(dNode)->Get_util ( Convert_OptInst_to_Inst(prevInst) );
		outfile << util << ", "; // needs to be comma separated
	}
	outfile.seekp( (long)outfile.tellp()-2 );	// delete the last ", "
	outfile << " }" << endl;
}
// *************************************************************************
// *************************************************************************
nfg_Vecs if_Gambit::Get_parent_values( BayesNet& bnet, const nfg_OptInst& optInst, 
												const nfg_Vertex& v )
{
	nfg_Vecs paVecs;
	//	Loop through each parent of v
	foreach ( nfg_Vertex vi, inv_adjacent_vertices(v, bnet.Get_graph()) )
	{
		// Make sure it is defined
		bassert ( optInst.at(vi) );

		//	add it to the list of parent vecs in order of vertex id
		paVecs.push_back( *optInst.at(vi) );
	}

	return paVecs;
}
// *************************************************************************
// *************************************************************************
if_Gambit::DiscreteSpaceIterator::DiscreteSpaceIterator( const nfg_Space& space )
{
	// Set the size based on the size of the space in each dimension
	foreach( nfg_Space1D space1d, space )
	{
		m_end.push_back( space1d.size() );
	}

	// The index should have the same size as m_size
	m_index.resize( m_end.size() );
}
// *************************************************************************
// *************************************************************************
void if_Gambit::DiscreteSpaceIterator::operator++( void )
{
	//	Do nothing if it is empty (this marks "past the end" condition)
	if ( m_index.empty() )
		return;

	for ( unsigned int ch = 0; ch < m_index.size(); ch++  )
	{
		//	Try to increment current index (starting from beginning)
		m_index.at(ch)++;
		//	Not at end of a channel
		if ( m_index.at(ch) != m_end.at(ch) )
		{
			break;
		}
		//	Reached the end of a channel, reset this channel
		//	and try to increase the next digit
		else
		{
			// reset channel
			m_index.at(ch) = 0;
			
			//	if this is the last channel, then mark as "past the end"
			if ( ch == m_index.size() - 1 )
			{
				m_index.clear();
				break;
			}
		}
	}
}
// *************************************************************************
// *************************************************************************
nfg_Vec if_Gambit::DiscreteSpaceIterator::operator()( void )
{
	// To be returned at the end
	nfg_Vec vec;

	// if it is "past the end", return an empty testcase
	if ( m_index.empty() )
		return vec;
	
	// return the current element that index is pointing to for each channel
	foreach( int ind, m_index )
	{
		vec.push_back( ind );
	}

	return vec;
}
// *************************************************************************
// *************************************************************************
void if_Gambit::Unit_subtest1( void )
{
	// Create a discrete space of 2, 3, 2
	nfg_Space space;
	space.push_back( Construct_discrete_space1d ( Vectorize(0,1) ));
	space.push_back( Construct_discrete_space1d ( Vectorize(0,1,2) ));
	space.push_back( Construct_discrete_space1d ( Vectorize(0,1) ));
	
	// Iterator that iterates the space based on the space
	DiscreteSpaceIterator it( space );
	while( !it().empty() )	// empty() is the flag for end state
	{
		// Make sure all combos are iterated through
		Print_vec( it() );
		++it;	// increment the iterator
	}
}

