
// *************************************************************************
//	LibNFG: Net-form game library
//
//	Author:		Ritchie Lee
//
//  Copyright 2011 United States Government as represented by the Administrator 
//  of the National Aeronautics and Space Administration.  All Rights Reserved.  
//  Licensed under the Apache License, Version 2.0 (the "License"); you may not 
//  use this file except in compliance with the License.  You may obtain a copy 
//  of the License at http://www.apache.org/licenses/LICENSE-2.0
//  Unless required by applicable law or agreed to in writing, software 
//  distributed under the License is distributed on an "AS IS" BASIS, WITHOUT 
//  WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the 
//  License for the specific language governing permissions and limitations 
//  under the License.
//
//	Revision Log
//
//	Username	Date		Description
//	-------------------------------------------------------
//	rlee		20110523	Created
//
// *************************************************************************

#ifndef INCLUDED_IF_MPI_H
#define INCLUDED_IF_MPI_H

#include "types.h"
#include "helper.h"

using namespace boost;

/**
 * if_Mpi: Interface to Boost.MPI for parallel computing capabilities.
 *		To use, you will need to install a version of MPI (e.g.,
 *		Microsoft Compute Cluster SDK, OpenMPI, MPICH2, etc.) and compile 
 *		the Boost.MPI library with it.
 *		For compatibility reasons, functions that refer to MPI are ifdef'd
 *		so that libnfg doesn't break if the user does not have Boost.MPI installed.
 *		To use Boost.MPI capability, add MPI_ENABLE to the preprocessor.
 */
class if_Mpi
{
public:

	/**	Constructor */
	if_Mpi( void );

	/**	Destructor */
	virtual ~ if_Mpi();

	/// Variables

	///	Methods
	/**
	 * Version 1: ifdef MPI_ENABLE
	 * Parallel_exec: Uses Boost.MPI to (embarassingly) parallelize the sampling.
	 *		This function establishes the structure, but it is the user that specifies 
	 *		the functions to be executed in the body.  The flow is as follows:
	 *		fo_parFunc() is called on all compute nodes, nfg_WtdInsts are gathered into
	 *		a vector at the head node, and passed to fo_postFunc(vector<nfg_WtdInsts>)
	 *		for post-processing on the head node.
	 *	fo_parFunc - function to be executed in parallel on all compute nodes
	 *	fo_postFunc - function to be executed on the head node on all gathered samples
	 **
	 * Version 2: ifndef MPI_ENABLE
	 * Parallel_exec: Preserves the functionality as above, but does not use MPI.  Tasks are
	 *		executed in a single thread.  This is only to maintain compatibility so that
	 *		libnfg will not break if you don't have mpi installed.
	 *	fo_parFunc - function to be executed in parallel on all compute nodes
	 *	fo_postFunc - function to be executed on the head node on all gathered samples
	 */
	static void Parallel_exec( const function<nfg_WtdInsts (void)>& fo_parFunc,
						const function<void (vector<nfg_WtdInsts>&)>& fo_postFunc );

	// Debug
	static void Unit_test( void );

private:

	/// Variables


	/// Methods
	/**
	 * Unit_test_parFunc: Support function for Unit_test that specifies
	 *		the function to be called during parallel processing.  This function will
	 *		be executed on all compute nodes.
	 *	returns: a list of weighted samples
	 */
	static nfg_WtdInsts		Unit_test_parFunc( void );
	/**
	 * Unit_test_postFunc: Support function for Unit_test that specifies
	 *		the function to be called during post processing. This function will only
	 *		execute on the head node
	 *	results - a vector of all the results gathered from Analysis2_parProc from all
	 *		compute nodes.
	 */
	static void				Unit_test_postFunc( const vector<nfg_WtdInsts>& results );

};

#endif

