
// *************************************************************************
//	LibNFG: Net-form game library
//
//	Author:		Ritchie Lee
//
//  Copyright 2011 United States Government as represented by the Administrator 
//  of the National Aeronautics and Space Administration.  All Rights Reserved.  
//  Licensed under the Apache License, Version 2.0 (the "License"); you may not 
//  use this file except in compliance with the License.  You may obtain a copy 
//  of the License at http://www.apache.org/licenses/LICENSE-2.0
//  Unless required by applicable law or agreed to in writing, software 
//  distributed under the License is distributed on an "AS IS" BASIS, WITHOUT 
//  WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the 
//  License for the specific language governing permissions and limitations 
//  under the License.
//
//	Revision Log
//
//	Username	Date		Description
//	-------------------------------------------------------
//	rlee		20101025	Created
//
// *************************************************************************

#ifndef INCLUDED_NDARRAY_H
#define INCLUDED_NDARRAY_H

#include <vector>
#include <assert.h>

using namespace std;

/**
* Class object used to contain and manipulate an n-dimensional array
*	 of doubles (i.e., arrays of arbitrary dimensions) n is not known a-priori.
*	Works by mapping tuples of indices into an array (memory space) with one 
*	dimensional indexing by remembering the needed strides.
*	
*/
class NDArray
{
public:

	/**	Constructor */
	NDArray( void );

	/**	Destructor */
	virtual ~ NDArray();

	/// Variables

	///	Methods
	/**
	 * Set_size: Specify the size of n-d array in each dimension
	 *	sizes - vector of ints, at(0) is the size of the n-d array in the 1st dimension
	 *		at(1) is the size of the n-d array in the 2nd dimension, etc.
	 */
	void			Set_size ( const vector<unsigned int>& sizes );
	/**
	 * Get_num_dims: Get number of dimension of n-d array
	 *	returns: number of dimensions
	 */
	unsigned int	Get_num_dims ( void ) { return m_ndims; };
	/**
	 * Get_size: Get size of n-d array in a particular dimension
	 *	dim - the dimension of interest
	 *	returns: size of array in dimension dimsizes
	 */	
	unsigned int	Get_size ( const int dim ) { return m_sizes.at(dim); };
	/**
	 * Set_value: Store a value into an index location.
	 *	idv: linear index location
	 *	val: value to be stored at the index location
	 */
	void			Set_value ( const unsigned int idv, const double val );
	/**
	 * Get_value: Get value at the index location.
	 *	idv: index location in tuple
	 *	returns: double stored at the index location
	 */	
	double			Get_value( const vector<unsigned int>& idv );
	/**
	 * Set_value: Store a value into an index location.
	 *	idv: tuple index location
	 *	val: value to be stored at the index location
	 */	
	void			Set_value ( const vector<unsigned int>& idv, const double val );
	/**
	 * Set_data: Store n-d array (already in 1D array format)
	 *	data: n-d array data in 1D array format
	 */	
	void			Set_data ( const vector<double>& data );
	/**
	 * Get_index: Converts a tuple index to a linear index
	 *	idv: tuple index location
	 *	returns: linear index location
	 */	
	unsigned int		Get_index ( const vector<unsigned int>& idv );
	

private:

	/// Variables
	unsigned int				m_ndims;		// number of dimensions
	unsigned int				m_elements;		// number of elements in total
	vector<unsigned int>		m_sizes;		// size in each dimension
	vector<unsigned int>		m_strides;		// used in the conversion formula for
												// tuple to linear indexing
	vector<double>				m_data;			// one dimensional array that stores all the data

	/// Methods
	/**
	 * Check_bounds: Check if tuple index is valid
	 *	idv: tuple index location
	 *	returns: true if index is within bounds of n-d array size, false otherwise
	 */	
	bool				Check_bounds( const vector<unsigned int>& idv );

};

#endif

