
// *************************************************************************
//	LibNFG: Net-form game library
//
//	Author:		Ritchie Lee
//
//  Copyright 2011 United States Government as represented by the Administrator 
//  of the National Aeronautics and Space Administration.  All Rights Reserved.  
//  Licensed under the Apache License, Version 2.0 (the "License"); you may not 
//  use this file except in compliance with the License.  You may obtain a copy 
//  of the License at http://www.apache.org/licenses/LICENSE-2.0
//  Unless required by applicable law or agreed to in writing, software 
//  distributed under the License is distributed on an "AS IS" BASIS, WITHOUT 
//  WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the 
//  License for the specific language governing permissions and limitations 
//  under the License.
//
//	Revision Log
//
//	Username	Date		Description
//	-------------------------------------------------------
//	rlee		20110329	Created
//
// *************************************************************************

#include "net_bobvsnature.h"

#include <boost/lambda/bind.hpp>
#include "bayesnet.h"
#include "alg_levelkrelaxed.h"
#include "cpd_uniform.h"
#include "cpd_decision.h"
#include "player.h"
#include "if_matlab.h"
#include "if_bayesnettoolbox.h"

using namespace nfg;
using namespace std;
using namespace boost;
using namespace boost::lambda;

// *************************************************************************
// *************************************************************************
void net_BobVsNature::Run ( void )
{
	//*********************************************
	///	 Set up Bayes Net

	// Specify vertex names
	string vertexNames[] = 
	{	"Bob", "Nature"	};

	// Specify vertex type
	nfg_eVertexType vertexTypes[] =
	{	DECISION, CHANCE };

	// Specify the space of each vertex
	nfg_Spaces spaces(N);
	spaces.at(D).push_back( nfg_Space1D(1, nfg_Range(0,5)) );
	spaces.at(C).push_back( nfg_Space1D(1, nfg_Range(0,5) ) );

	////	Make Bayes Net
	// Pass all specified values above to the bnet constructor
	BayesNet bnet( "Bob vs Nature", N,
		nfg_Edges(),			// no edges
		nfg_Dim(N,1),			// all sizes are 1
		spaces,
		nfg_Mask(N, true),		// all continuous
		nfg_eVertexTypes(vertexTypes,vertexTypes+N),
		nfg_Mask(N, false),		// no branch nodes
		nfg_Mask(N, false) );	// no dno nodes

	// Set vertex names
	bnet.Set_vertex_names ( vector<string>(vertexNames,vertexNames+N) );

	//	Declare players
	Player P1 ( bind(&Util_function_P1, _1) );		// pass in utility function
	P1.Add_controlled_node(D); // Specify which decision nodes are controlled by which players

	// Associate players with bnet
	bnet.Add_player(P1);

	//	Create cpds
	// S is uniform cpd
	cpd_Uniform		C_cpd ( nfg_Vec(1,0.0), nfg_Vec(1,5.0) );	// uniform between 0 and 5
	cpd_Decision	D_cpd ( P1 );								// associate with player

	//	Link cpds to bnet
	bnet.Set_cpd(C, &C_cpd );
	bnet.Set_cpd(D, &D_cpd );

	//	Error check
	bassert ( bnet.Check() );

	//**************************************************
	///	Run main program
	
	// Draw graph (to make sure the code understood your graph correctly)
	if_BayesNetToolbox::Draw_graph( "Output\\BvN_draw.m", bnet );

	// Run custom analysis on bnet
	Analysis1( bnet );

	return;
}
// *************************************************************************
// *************************************************************************
double net_BobVsNature::Util_function_P1 ( const nfg_Inst& instVecs )
{
	//	Difference between Bob and Nature's move
	return instVecs.at(D).at(0) - instVecs.at(C).at(0);
}
// *************************************************************************
// *************************************************************************
nfg_Vec net_BobVsNature::Sample_movespace_D ( const nfg_Vecs& paVecs )
{
	//	The parent vectors are ignored

	// Uniform between 0 and 5
	return nfg_Vec(1, Rand_uniform_range(0,5) );
}
// *************************************************************************
// *************************************************************************
nfg_Vec net_BobVsNature::Sample_level0_D ( const nfg_Vecs& paVecs )
{
	//	The parent vectors are ignored

	// Uniform between 0 and 5
	return nfg_Vec(1, Rand_uniform_range(0,5) );
}
// *************************************************************************
// *************************************************************************
void net_BobVsNature::Analysis1 ( BayesNet& bnet )
{
	// Declare inference engine
	alg_LevelKRelaxed engine;
	
	// Specify level0 distributions (functions) for each decision node
	nfg_SampleFuncs fo_Sample_level0_inst(N);
	fo_Sample_level0_inst.at(D) = bind(&Sample_level0_D, _1);

	// Specify satisficing distributions (functions) for each decision node
	nfg_SampleFuncs fo_Sample_satisf_inst(N);
	fo_Sample_satisf_inst.at(D) = bind(&Sample_movespace_D,_1);	// specify satisficing distr.

	// Set level K parameters
	engine.Set_params( bnet, alg_LevelKRelaxed::D_RELAXED, 5, 10, fo_Sample_level0_inst,
								fo_Sample_satisf_inst,
								alg_LevelKRelaxed::BEST, true );
	
	// Used to contain the samples
	nfg_WtdInsts wInsts;

	// Collect 500 samples of the net
	for ( int i = 0; i < 500; i++ )
	{
		cout << i << endl;		// Progress indicator to console
		// Generate a sample and add it to the list
		wInsts.push_back( engine.LW_forward_sample( bnet, 1 ) );
	}
	// Output the samples to a Matlab file for post-processing
	if_Matlab::Export_WtdInsts_to_file ( "Output\\BvN_vars.m", "samples", wInsts, false );

	//	Track Bob's utility distribution
	nfg_Vec P1Utils;
	foreach ( nfg_WtdInst wInst, wInsts )
	{
		// Map each instantiation (sample) of the net to the player's utilities
		// and save them to a list
		P1Utils.push_back( Util_function_P1(wInst.m_inst) );
	}
	// Output the utility values to a Matlab file for post-processing
	if_Matlab::Export_Vec_to_file ( "Output\\BvN_vars.m", "P1Utils", P1Utils, true );
}



