
// *************************************************************************
//	LibNFG: Net-form game library
//
//	Author:		Ritchie Lee
//
//  Copyright 2011 United States Government as represented by the Administrator 
//  of the National Aeronautics and Space Administration.  All Rights Reserved.  
//  Licensed under the Apache License, Version 2.0 (the "License"); you may not 
//  use this file except in compliance with the License.  You may obtain a copy 
//  of the License at http://www.apache.org/licenses/LICENSE-2.0
//  Unless required by applicable law or agreed to in writing, software 
//  distributed under the License is distributed on an "AS IS" BASIS, WITHOUT 
//  WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the 
//  License for the specific language governing permissions and limitations 
//  under the License.
//
//	Revision Log
//
//	Username	Date		Description
//	-------------------------------------------------------
//	rlee		20110217	Created
//
// *************************************************************************

#include "net_mergeandspace.h"

#include <boost/lambda/bind.hpp>
#include "bayesnet.h"
#include "alg_levelkrelaxed.h"
#include "cpd.h"
#include "cpd_uniform.h"
#include "cpd_gaussian.h"
#include "cpd_decision.h"
#include "player.h"
#include "if_matlab.h"
#include "if_bayesnettoolbox.h"

using namespace nfg;
using namespace std;
using namespace boost;
using namespace boost::lambda;

// *************************************************************************
// *************************************************************************
void net_MergeAndSpace::Run ( void )
{
	//*********************************************
	///	 Set up Bayes Net

	// Specify vertex names
	string vertexNames[] = 
	{	"S1", "S2", "W1", "W2", "D1", "D2"	};

	// Define Edges of the graph
	nfg_Edge edgeArray[] = 
	{ 
		nfg_Edge(S1,D1), nfg_Edge(S1,W2), 
		nfg_Edge(S2,D2), nfg_Edge(S2,W1),
		nfg_Edge(W1,D1), nfg_Edge(W2,D2) 
	};

	// Specify vertex type
	nfg_eVertexType vertexTypes[] =
	{	CHANCE, CHANCE, CHANCE, CHANCE, DECISION, DECISION	};

	// Specify the space of each vertex
	nfg_Spaces spaces(N);
	spaces.at(S1).push_back( nfg_Space1D(1, nfg_Range(0,5) ));
	spaces.at(S2).push_back( nfg_Space1D(1, nfg_Range(0,5) ));
	spaces.at(D1).push_back( nfg_Space1D(1, nfg_Range(0,5) ));
	spaces.at(D2).push_back( nfg_Space1D(1, nfg_Range(0,5) ));
	// Note: no constraints on W1, W2 because they are Gaussian

	////	Make Bayes Net
	// Pass all specified values above to the bnet constructor
	BayesNet bnet( "Merging and Spacing", N,
		nfg_Edges(edgeArray,edgeArray+sizeof(edgeArray)/sizeof(nfg_Edge)), 
		nfg_Dim(N,1),		// all nodes have size 1
		spaces,
		nfg_Mask(N,true),	// all continuous
		nfg_eVertexTypes(vertexTypes,vertexTypes+N),
		nfg_Mask(N, false),		// no branch nodes
		nfg_Mask(N, false));	// no nodes with dno
	// Set vertex names
	bnet.Set_vertex_names ( vector<string>(vertexNames,vertexNames+N) );

	//	Declare players
	Player P1 ( bind(&Util_function_P1, _1) );	// pass in utility function
	Player P2 ( bind(&Util_function_P2, _1) );	// pass in utility function
	// Specify which decision nodes are controlled by which players
	P1.Add_controlled_node(D1);
	P2.Add_controlled_node(D2);

	// Associate players with bnet
	bnet.Add_player(P1);
	bnet.Add_player(P2);

	//	Create cpds
	// S1, S2 are uniform on different ranges
	cpd_Uniform		S1_cpd ( nfg_Vec(1,2.0), nfg_Vec(1,5.0) );
	cpd_Uniform		S2_cpd ( nfg_Vec(1,0.0), nfg_Vec(1,3.0) );
	// W1 and W2 are observations on either S1 and S2
	cpd_Gaussian	W1_cpd ( nfg_Vec(1,0.5) );
	cpd_Gaussian	W2_cpd ( nfg_Vec(1,0.5) );
	// Decision nodes
	cpd_Decision	D1_cpd ( P1 );	// associate with player
	cpd_Decision	D2_cpd ( P2 );	// associate with player

	//	Link cpds to bnet
	bnet.Set_cpd(S1, &S1_cpd );
	bnet.Set_cpd(S2, &S2_cpd );
	bnet.Set_cpd(W1, &W1_cpd );
	bnet.Set_cpd(W2, &W2_cpd );
	bnet.Set_cpd(D1, &D1_cpd );
	bnet.Set_cpd(D2, &D2_cpd );

	//	Error check
	bassert ( bnet.Check() );

	//**************************************************
	///	Run main program
	
	// Draw graph (to make sure the code understood your graph correctly)
	if_BayesNetToolbox::Draw_graph( "Output\\MnS_draw.m", bnet );

	// Run custom analysis on bnet
	Analysis1(bnet);

	return;
}
// *************************************************************************
// *************************************************************************
double net_MergeAndSpace::Util_function_P1 ( const nfg_Inst& instVecs )
{

	//	weights
	double a1 = 0.8;
	double a2 = 1 - a1;

	//	Term 1: How close to the other player, higher is better
	double u1 = abs ( instVecs.at(D2).at(0) - instVecs.at(D1).at(0));

	//	Term 2: How much effort was spent to move, lower is better
	double u2 = abs ( instVecs.at(D1).at(0) - instVecs.at(S1).at(0) );

	// linear combo of all terms
	return a1*u1 - a2*u2;
}
// *************************************************************************
// *************************************************************************
double net_MergeAndSpace::Util_function_P2 ( const nfg_Inst& instVecs )
{

	// weights
	double a1 = 0.8;
	double a2 = 1 - a1;

	//	Term 1: How close to the other player, higher is better
	double u1 = abs ( instVecs.at(D2).at(0) - instVecs.at(D1).at(0));

	//	Term 2: How much effort was spent to move, lower is better
	double u2 = abs ( instVecs.at(D2).at(0) - instVecs.at(S2).at(0) );

	// linear combo of all terms
	return a1*u1 - a2*u2;
}
// *************************************************************************
// *************************************************************************
nfg_Vec net_MergeAndSpace::Sample_movespace ( const nfg_Vecs& paVecs )
{
	//	The parent vectors are ignored

	// Uniform on 0 to 5
	return nfg_Vec(1, Rand_uniform_range(0,5) );
}
// *************************************************************************
// *************************************************************************
nfg_Vec net_MergeAndSpace::Sample_level0 ( const nfg_Vecs& paVecs )
{
	//	The parent vectors are ignored

	// Uniform on 0 to 5
	return nfg_Vec(1, Rand_uniform_range(0,5) );
}
// *************************************************************************
// *************************************************************************
void net_MergeAndSpace::Analysis1 ( BayesNet& bnet )
{
	// Declare inference engine
	alg_LevelKRelaxed engine;

	// Specify level0 distributions (functions) for each decision node
	nfg_SampleFuncs fo_Sample_level0_inst(N);
	fo_Sample_level0_inst.at(D1) = bind(&Sample_level0, _1);	// specify level0 distr.
	fo_Sample_level0_inst.at(D2) = bind(&Sample_level0, _1);	// specify level0 distr.

	// Specify satisficing distributions (functions) for each decision node
	nfg_SampleFuncs fo_Sample_satisf_inst(N);
	fo_Sample_satisf_inst.at(D1) = bind(&Sample_movespace,_1);	// specify satisficing distr.
	fo_Sample_satisf_inst.at(D2) = bind(&Sample_movespace,_1);	// specify satisficing distr.

	// Set level K parameters
	engine.Set_params( bnet, alg_LevelKRelaxed::D_RELAXED, 5, 10, fo_Sample_level0_inst,
								fo_Sample_satisf_inst,
								alg_LevelKRelaxed::BEST, true );
	
	// Used to contain the samples
	nfg_WtdInsts wInsts;

	// Collect 500 samples of the net
	for ( int i = 0; i < 500; i++ )
	{
		cout << i << endl;		// Progress indicator to console
		// Generate a sample and add it to the list
		wInsts.push_back( engine.LW_forward_sample( bnet, 2 ) );
	}
	// Output the samples to a Matlab file for post-processing
	if_Matlab::Export_WtdInsts_to_file ( "Output\\MnS_vars.m", "samples", wInsts, false );

	//	Track players' utility distribution
	nfg_Vec P1Utils, P2Utils;
	foreach ( nfg_WtdInst wInst, wInsts )
	{
		// Map each instantiation (sample) of the net to the player's utilities
		// and save them to a list
		P1Utils.push_back( Util_function_P1(wInst.m_inst) );
		P2Utils.push_back( Util_function_P2(wInst.m_inst) );
	}
	// Output the utility values to a Matlab file for post-processing
	if_Matlab::Export_Vec_to_file ( "Output\\MnS_vars.m", "P1Utils", P1Utils, true );
	if_Matlab::Export_Vec_to_file ( "Output\\MnS_vars.m", "P2Utils", P2Utils, true );
}

