
// *************************************************************************
//	LibNFG: Net-form game library
//
//	Author:		Ritchie Lee
//
//  Copyright 2011 United States Government as represented by the Administrator 
//  of the National Aeronautics and Space Administration.  All Rights Reserved.  
//  Licensed under the Apache License, Version 2.0 (the "License"); you may not 
//  use this file except in compliance with the License.  You may obtain a copy 
//  of the License at http://www.apache.org/licenses/LICENSE-2.0
//  Unless required by applicable law or agreed to in writing, software 
//  distributed under the License is distributed on an "AS IS" BASIS, WITHOUT 
//  WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the 
//  License for the specific language governing permissions and limitations 
//  under the License.
//
//	Revision Log
//
//	Username	Date		Description
//	-------------------------------------------------------
//	rlee		20110217	Created
//
// *************************************************************************

#include "net_mergeandspace2step.h"

#include <boost/lambda/bind.hpp>
#include "bayesnet.h"
#include "alg_levelkrelaxed.h"
#include "cpd.h"
#include "cpd_uniform.h"
#include "cpd_gaussian.h"
#include "cpd_decision.h"
#include "player.h"
#include "if_matlab.h"
#include "if_bayesnettoolbox.h"
#include "if_mpi.h"

using namespace nfg;
using namespace std;
using namespace boost;
using namespace boost::lambda;

// *************************************************************************
// *************************************************************************
void net_MergeAndSpace2Step::Run ( void )
{
	//*********************************************
	///	 Set up Bayes Net
	//	This is the two-stage version of MergeAndSpace

	// Specify vertex names
	string vertexNames[] = 
	{	"S1", "S2", "W1", "W2", "D1", "D2", "W3", "W4", "D3", "D4"	};

	// Define Edges of the graph
	nfg_Edge edgeArray[] = 
	{ 
		nfg_Edge(S1,D1), nfg_Edge(S1,W2), 
		nfg_Edge(S2,D2), nfg_Edge(S2,W1),
		nfg_Edge(W1,D1), nfg_Edge(W2,D2),
		nfg_Edge(D1,W4), nfg_Edge(D2,W3),	// Begin second stage
		nfg_Edge(D1,D3), nfg_Edge(D2,D4),	// Players remember what they did before
		nfg_Edge(W3,D3), nfg_Edge(W4,D4),
		nfg_Edge(W1,D3), nfg_Edge(W2,D4)	// Players remember what they saw before
	};


	// Specify vertex type
	nfg_eVertexType vertexTypes[] =
	{	CHANCE, CHANCE, CHANCE, CHANCE, DECISION, DECISION, CHANCE, CHANCE, DECISION, DECISION	};

	// Specify the space of each vertex
	nfg_Spaces spaces(N);
	spaces.at(S1).push_back( nfg_Space1D(1, nfg_Range(2,5) ));
	spaces.at(S2).push_back( nfg_Space1D(1, nfg_Range(0,3) ));
	spaces.at(D1).push_back( nfg_Space1D(1, nfg_Range(0,5) ));
	spaces.at(D2).push_back( nfg_Space1D(1, nfg_Range(0,5) ));
	spaces.at(D3).push_back( nfg_Space1D(1, nfg_Range(0,5) ));
	spaces.at(D4).push_back( nfg_Space1D(1, nfg_Range(0,5) ));
	// Note: no constraints on W's because they are Gaussian

	////	Make Bayes Net
	// Pass all specified values above to the bnet constructor
	BayesNet bnet( "Merging and Spacing - 2 time step", N,
		nfg_Edges(edgeArray,edgeArray+sizeof(edgeArray)/sizeof(nfg_Edge)), 
		nfg_Dim(N,1),			// all sizes are 1
		spaces,
		nfg_Mask(N, true),		// all nodes are continuous
		nfg_eVertexTypes(vertexTypes,vertexTypes+N),
		nfg_Mask(N, false),		// no branching
		nfg_Mask(N, false));	// no nodes with dno
	// Set vertex names
	bnet.Set_vertex_names ( vector<string>(vertexNames,vertexNames+N) );

	//	Declare players
	Player P1 ( bind(&Util_function_P1, _1), "Player1" );	// pass in utility function
	Player P2 ( bind(&Util_function_P2, _1), "Player2" );	// pass in utility function
	// Specify which decision nodes are controlled by which players
	P1.Add_controlled_node(D1);
	P1.Add_controlled_node(D3);
	P2.Add_controlled_node(D2);
	P2.Add_controlled_node(D4);

	// Associate players with bnet
	bnet.Add_player(P1);
	bnet.Add_player(P2);

	//	Create cpds
	// S1, S2 are uniform on different ranges
	cpd_Uniform		S1_cpd ( nfg_Vec(1,2.0), nfg_Vec(1,5.0) );
	cpd_Uniform		S2_cpd ( nfg_Vec(1,0.0), nfg_Vec(1,3.0) );
	// W1 and W2 are observations on either S1 and S2
	cpd_Gaussian	W1_cpd ( nfg_Vec(1,0.5) );
	cpd_Gaussian	W2_cpd ( nfg_Vec(1,0.5) );
	// 1st time step decision nodes
	cpd_Decision	D1_cpd ( P1 );	// pass in satisficing distr.
	cpd_Decision	D2_cpd ( P2 );	// pass in satisficing distr.
	// W3 and W4 are observations on either D1 and D2
	cpd_Gaussian	W3_cpd ( nfg_Vec(1,0.5) );
	cpd_Gaussian	W4_cpd ( nfg_Vec(1,0.5) );
	// 2nd step decision nodes
	cpd_Decision	D3_cpd ( P1 );	// pass in satisficing distr.
	cpd_Decision	D4_cpd ( P2 );	// pass in satisficing distr.

	//	Link cpds to bnet using automatic linking
	bnet.Set_cpd(S1, &S1_cpd );
	bnet.Set_cpd(S2, &S2_cpd );
	bnet.Set_cpd(W1, &W1_cpd );
	bnet.Set_cpd(W2, &W2_cpd );
	bnet.Set_cpd(D1, &D1_cpd );
	bnet.Set_cpd(D2, &D2_cpd );
	bnet.Set_cpd(W3, &W3_cpd );
	bnet.Set_cpd(W4, &W4_cpd );
	bnet.Set_cpd(D3, &D3_cpd );
	bnet.Set_cpd(D4, &D4_cpd );

	//	Error check
	bassert ( bnet.Check() );

	//**************************************************
	///	Run main program
	
	// Draw graph (to make sure the code understood your graph correctly)
	if_BayesNetToolbox::Draw_graph( "Output\\MnS2_draw.m", bnet );

	// Choose custom analysis to run on bnet
	Analysis1(bnet);
	//Analysis2(bnet);

	return;
}
// *************************************************************************
// *************************************************************************
double net_MergeAndSpace2Step::Util_function_P1 ( const nfg_Inst& instVecs )
{
	// Weights
	double a1 = 0.5;
	double a2 = 0.3;
	double a3 = 1.0 - a1 - a2;

	//	Term 1: How close to the other player, final position, higher is better
	double u1 = abs ( instVecs.at(D4).at(0) - instVecs.at(D3).at(0));

	//	Term 2: How close to the other player, intermediate position, higher is better
	double u2 = abs ( instVecs.at(D2).at(0) - instVecs.at(D1).at(0));

	//	Term 2: How much effort was spent to move, lower is better
	double u3 = abs ( instVecs.at(D1).at(0) - instVecs.at(S1).at(0) ) +
					abs ( instVecs.at(D3).at(0) - instVecs.at(D1).at(0) );


	// linear combo of all terms
	return a1*u1 + a2*u2 - a3*u3;
}
// *************************************************************************
// *************************************************************************
double net_MergeAndSpace2Step::Util_function_P2 ( const nfg_Inst& instVecs )
{
	// Weights
	double a1 = 0.5;
	double a2 = 0.3;
	double a3 = 1.0 - a1 - a2;

	//	Term 1: How close to the other player, final position, higher is better
	double u1 = abs ( instVecs.at(D4).at(0) - instVecs.at(D3).at(0));

	//	Term 2: How close to the other player, intermediate position, higher is better
	double u2 = abs ( instVecs.at(D2).at(0) - instVecs.at(D1).at(0));

	//	Term 2: How much effort was spent to move, lower is better
	double u3 = abs ( instVecs.at(D2).at(0) - instVecs.at(S2).at(0) ) +
					abs ( instVecs.at(D4).at(0) - instVecs.at(D2).at(0) );

	// linear combo of all terms
	return a1*u1 + a2*u2 - a3*u3;
}
// *************************************************************************
// *************************************************************************
nfg_Vec net_MergeAndSpace2Step::Sample_movespace ( const nfg_Vecs& paVecs )
{
	//	The parent vectors are ignored

	// Uniform on 0 to 5
	return nfg_Vec(1, Rand_uniform_range(0,5) );
}
// *************************************************************************
// *************************************************************************
nfg_Vec net_MergeAndSpace2Step::Sample_level0 ( const nfg_Vecs& paVecs )
{
	//	The parent vectors are ignored

	// Uniform on 0 to 5
	return nfg_Vec(1, Rand_uniform_range(0,5) );
}
// *************************************************************************
// *************************************************************************
void net_MergeAndSpace2Step::Analysis1 ( BayesNet& bnet )
{
	// Declare inference engine
	alg_LevelKRelaxed engine;

	// Specify level0 distributions (functions) for each decision node
	nfg_SampleFuncs fo_Sample_level0_inst(N);
	fo_Sample_level0_inst.at(D1) = bind(&Sample_level0, _1);	// specify level0 distr.
	fo_Sample_level0_inst.at(D2) = bind(&Sample_level0, _1);	// specify level0 distr.
	fo_Sample_level0_inst.at(D3) = bind(&Sample_level0, _1);	// specify level0 distr.
	fo_Sample_level0_inst.at(D4) = bind(&Sample_level0, _1);	// specify level0 distr.

	// Specify satisficing distributions (functions) for each decision node
	nfg_SampleFuncs fo_Sample_satisf_inst(N);
	fo_Sample_satisf_inst.at(D1) = bind(&Sample_movespace,_1);	// specify satisficing distr.
	fo_Sample_satisf_inst.at(D2) = bind(&Sample_movespace,_1);	// specify satisficing distr.
	fo_Sample_satisf_inst.at(D3) = bind(&Sample_movespace,_1);	// specify satisficing distr.
	fo_Sample_satisf_inst.at(D4) = bind(&Sample_movespace,_1);	// specify satisficing distr.

	// Set level K parameters
	engine.Set_params( bnet, alg_LevelKRelaxed::D_RELAXED, 5, 10, fo_Sample_level0_inst,
								fo_Sample_satisf_inst,
								alg_LevelKRelaxed::BEST, true );
	
	// Used to contain the samples
	nfg_WtdInsts wInsts;

	// Collect 500 samples of the net
	for ( int i = 0; i < 500; i++ )
	{
		cout << i << endl;		// Progress indicator to console
		// Generate a sample and add it to the list
		wInsts.push_back( engine.LW_forward_sample( bnet, 1 ) );
	}
	// Output the samples to a Matlab file for post-processing
	if_Matlab::Export_WtdInsts_to_file ( "Output\\MnS2_vars.m", "samples", wInsts, false );

	//	Track players' utility distribution
	nfg_Vec P1Utils, P2Utils;
	foreach ( nfg_WtdInst wInst, wInsts )
	{
		// Map each instantiation (sample) of the net to the player's utilities
		// and save them to a list
		P1Utils.push_back( Util_function_P1(wInst.m_inst) );
		P2Utils.push_back( Util_function_P2(wInst.m_inst) );
	}
	// Output the utility values to a Matlab file for post-processing
	if_Matlab::Export_Vec_to_file ( "Output\\MnS2_vars.m", "P1Utils", P1Utils, true );
	if_Matlab::Export_Vec_to_file ( "Output\\MnS2_vars.m", "P2Utils", P2Utils, true );
}
// *************************************************************************
// *************************************************************************
void net_MergeAndSpace2Step::Analysis2 ( BayesNet& bnet )
{
	cout << "Running MnS2::Analysis2" << endl;

	// Declare inference engine
	alg_LevelKRelaxed engine;

	// Specify level0 distributions (functions) for each decision node
	nfg_SampleFuncs fo_Sample_level0_inst(N);
	fo_Sample_level0_inst.at(D1) = bind(&Sample_level0, _1);	// specify level0 distr.
	fo_Sample_level0_inst.at(D2) = bind(&Sample_level0, _1);	// specify level0 distr.
	fo_Sample_level0_inst.at(D3) = bind(&Sample_level0, _1);	// specify level0 distr.
	fo_Sample_level0_inst.at(D4) = bind(&Sample_level0, _1);	// specify level0 distr.

	// Specify satisficing distributions (functions) for each decision node
	nfg_SampleFuncs fo_Sample_satisf_inst(N);
	fo_Sample_satisf_inst.at(D1) = bind(&Sample_movespace,_1);	// specify satisficing distr.
	fo_Sample_satisf_inst.at(D2) = bind(&Sample_movespace,_1);	// specify satisficing distr.
	fo_Sample_satisf_inst.at(D3) = bind(&Sample_movespace,_1);	// specify satisficing distr.
	fo_Sample_satisf_inst.at(D4) = bind(&Sample_movespace,_1);	// specify satisficing distr.

	// Set level K parameters
	engine.Set_params( bnet, alg_LevelKRelaxed::D_RELAXED, 5, 10, fo_Sample_level0_inst,
								fo_Sample_satisf_inst,
								alg_LevelKRelaxed::BEST, true );
	
	// Collect some samples from each processor using parallel processing module
	if_Mpi::Parallel_exec( 
		bind( &Analysis2_parProc, ref(engine), ref(bnet) ), 
		bind( &Analysis2_postProc,_1) );

	
}
// *************************************************************************
// *************************************************************************
void net_MergeAndSpace2Step::Analysis2_postProc ( const vector<nfg_WtdInsts>& wInstsVec  )
{
	nfg_WtdInsts finalList;

	// Unpack the lists and reconstruct a single large list
	foreach( const nfg_WtdInsts& wInsts, wInstsVec )
	{
		finalList.insert( finalList.end(), wInsts.begin(), wInsts.end() );
	}
	
	// Output the samples to a Matlab file for post-processing
	if_Matlab::Export_WtdInsts_to_file ( "Output\\MnS2_vars.m", "samples", finalList, false );

	//	Track players' utility distribution
	nfg_Vec P1Utils, P2Utils;
	foreach ( nfg_WtdInst wInst, finalList )
	{
		// Map each instantiation (sample) of the net to the player's utilities
		// and save them to a list
		P1Utils.push_back( Util_function_P1(wInst.m_inst) );
		P2Utils.push_back( Util_function_P2(wInst.m_inst) );
	}
	// Output the utility values to a Matlab file for post-processing
	if_Matlab::Export_Vec_to_file ( "Output\\MnS2_vars.m", "P1Utils", P1Utils, true );
	if_Matlab::Export_Vec_to_file ( "Output\\MnS2_vars.m", "P2Utils", P2Utils, true );
}
// *************************************************************************
// *************************************************************************
nfg_WtdInsts net_MergeAndSpace2Step::Analysis2_parProc ( alg_LevelKRelaxed& engine, BayesNet& bnet  )
{
	nfg_WtdInsts wInsts;

	// number of samples on each processor
	int numSamples = 20;
	for ( int i = 0; i < numSamples; i++ )
	{
		cout << i << endl;		// Progress indicator to console
		// Generate a sample and add it to the list
		wInsts.push_back( engine.LW_forward_sample( bnet, 1 ) );
	}

	return wInsts;
}


