
// *************************************************************************
//	LibNFG: Net-form game library
//
//	Author:		Ritchie Lee
//
//  Copyright 2011 United States Government as represented by the Administrator 
//  of the National Aeronautics and Space Administration.  All Rights Reserved.  
//  Licensed under the Apache License, Version 2.0 (the "License"); you may not 
//  use this file except in compliance with the License.  You may obtain a copy 
//  of the License at http://www.apache.org/licenses/LICENSE-2.0
//  Unless required by applicable law or agreed to in writing, software 
//  distributed under the License is distributed on an "AS IS" BASIS, WITHOUT 
//  WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the 
//  License for the specific language governing permissions and limitations 
//  under the License.
//
//	Revision Log
//
//	Username	Date		Description
//	-------------------------------------------------------
//	rlee		20110329	Created
//
// *************************************************************************

#include "net_prisonersdilemma.h"

#include <boost/lambda/bind.hpp>
#include "bayesnet.h"
#include "alg_levelkrelaxed.h"
#include "cpd.h"
#include "cpd_decision.h"
#include "player.h"
#include "if_matlab.h"
#include "if_bayesnettoolbox.h"
#include "if_gambit.h"

using namespace nfg;
using namespace std;
using namespace boost;
using namespace boost::lambda;

// *************************************************************************
// *************************************************************************
void net_PrisonersDilemma::Run ( void )
{
	//*********************************************
	///	 Set up Bayes Net

	// Specify vertex names
	string vertexNames[] = 
	{	"Row", "Column"	};

	// Specify the space of each vertex
	nfg_Spaces spaces(N);
	spaces.at(D1).push_back( Construct_discrete_space1d ( Vectorize(0,1) ));
	spaces.at(D2).push_back( Construct_discrete_space1d ( Vectorize(0,1) ));

	////	Make Bayes Net
	// Pass all specified values above to the bnet constructor
	BayesNet bnet( "Prisoner's Dilemma", N,
		nfg_Edges(),		// no edges 
		nfg_Dim(N,1),		// all sizes are 1
		spaces,
		nfg_Mask(N,false),	// all discrete
		nfg_eVertexTypes(N,DECISION), // all decision nodes
		nfg_Mask(N, false),		// no branch nodes
		nfg_Mask(N, false));	// no nodes with dno
	// Set vertex names
	bnet.Set_vertex_names ( vector<string>(vertexNames,vertexNames+N) );

	//	Declare players
	Player P1 ( bind(&Util_function_P1, _1) );	// pass in utility function
	Player P2 ( bind(&Util_function_P2, _1) );	// pass in utility function
	// Specify which decision nodes are controlled by which players
	P1.Add_controlled_node(D1);
	P2.Add_controlled_node(D2);

	// Associate players with bnet
	bnet.Add_player(P1);
	bnet.Add_player(P2);

	//	Create cpds
	// Both are decision nodes
	cpd_Decision	D1_cpd ( P1 );	// Associate decision node with player
	cpd_Decision	D2_cpd ( P2 );	// Associate decision node with player

	//	Link cpds to bnet using automatic linking
	bnet.Set_cpd(D1, &D1_cpd );
	bnet.Set_cpd(D2, &D2_cpd );

	//	Error check
	bassert ( bnet.Check() );

	//**************************************************
	///	Run main program

	// Draw graph (to make sure the code understood your graph correctly)
	if_BayesNetToolbox::Draw_graph( "Output\\PD_draw.m", bnet );

	// Run custom analysis on bnet
	Analysis1(bnet);
	//Analysis2(bnet);

	return;
}
// *************************************************************************
// *************************************************************************
double net_PrisonersDilemma::Util_function_P1 ( const nfg_Inst& instVecs )
{
	//	0 is cooperate, 1 is defect
	//	According to payout table.
	//		C		D
	//	C [ (3,3), (0,5) ]
	//	C [ (5,0), (1,1) ]

	double payout = 0.0;
	
	//	Cooperate-cooperate
	if ( instVecs.at(D1).at(0) == 0 && instVecs.at(D2).at(0) == 0 )
		payout = 3;
	//	Cooperate-defect
	else if ( instVecs.at(D1).at(0) == 0 && instVecs.at(D2).at(0) == 1 )
		payout = 0;
	//	Defect-cooperate
	else if ( instVecs.at(D1).at(0) == 1 && instVecs.at(D2).at(0) == 0 )
		payout = 5;
	//	Defect-Defect
	else if ( instVecs.at(D1).at(0) == 1 && instVecs.at(D2).at(0) == 1 )
		payout = 1;
	//	Something's wrong...
	else
		bassert(false);

	return payout;
}
// *************************************************************************
// *************************************************************************
double net_PrisonersDilemma::Util_function_P2 ( const nfg_Inst& instVecs )
{
	//	0 is cooperate, 1 is defect
	//	According to payout table.
	//		C		D
	//	C [ (3,3), (0,5) ]
	//	C [ (5,0), (1,1) ]

	double payout = 0.0;
	
	//	Cooperate-cooperate
	if ( instVecs.at(D1).at(0) == 0 && instVecs.at(D2).at(0) == 0 )
		payout = 3;
	//	Cooperate-defect
	else if ( instVecs.at(D1).at(0) == 0 && instVecs.at(D2).at(0) == 1 )
		payout = 5;
	//	Defect-cooperate
	else if ( instVecs.at(D1).at(0) == 1 && instVecs.at(D2).at(0) == 0 )
		payout = 0;
	//	Defect-Defect
	else if ( instVecs.at(D1).at(0) == 1 && instVecs.at(D2).at(0) == 1 )
		payout = 1;
	//	Something's wrong...
	else
		bassert(false);

	return payout;
}
// *************************************************************************
// *************************************************************************
nfg_Vec net_PrisonersDilemma::Sample_level0 ( const nfg_Vecs& paVecs )
{
	//	The parent vectors are ignored

	//	uniform distribution over the 2 moves
	return nfg_Vec(1, (double)Rand_int(0,1) );
}
// *************************************************************************
// *************************************************************************
nfg_Vec net_PrisonersDilemma::Sample_movespace ( const nfg_Vecs& paVecs )
{
	//	The parent vectors are ignored

	//	Uniform over 2 possible moves
	return nfg_Vec(1, (double)Rand_int(0,1) );
}
// *************************************************************************
// *************************************************************************
void net_PrisonersDilemma::Analysis1 ( BayesNet& bnet )
{
	cout << "Running Analysis1..." << endl;

	// Declare inference engine
	alg_LevelKRelaxed engine;

	// Specify level0 distributions (functions) for each decision node
	nfg_SampleFuncs fo_Sample_level0_inst(N);
	fo_Sample_level0_inst.at(D1) = bind(&Sample_level0, _1);	// specify level0 distr.
	fo_Sample_level0_inst.at(D2) = bind(&Sample_level0, _1);	// specify level0 distr.

	// Specify satisficing distributions (functions) for each decision node
	nfg_SampleFuncs fo_Sample_satisf_inst(N);
	fo_Sample_satisf_inst.at(D1) = bind(&Sample_movespace,_1);	// specify satisficing distr.
	fo_Sample_satisf_inst.at(D2) = bind(&Sample_movespace,_1);	// specify satisficing distr.

	// Set level K parameters
	engine.Set_params( bnet, alg_LevelKRelaxed::D_RELAXED, 5, 10, fo_Sample_level0_inst,
								fo_Sample_satisf_inst,
								alg_LevelKRelaxed::BEST, true );
	
	// Used to contain the samples
	nfg_WtdInsts wInsts;

	// Collect 500 samples of the net
	for ( int i = 0; i < 500; i++ )
	{
		cout << i << endl;		// Progress indicator to console
		// Generate a sample and add it to the list
		wInsts.push_back( engine.LW_forward_sample( bnet, 2 ) );
	}
	// Output the samples to a Matlab file for post-processing
	if_Matlab::Export_WtdInsts_to_file ( "Output\\PD_vars.m", "samples", wInsts, false );

	//	Track P1 and P2's utility distributions
	nfg_Vec P1Utils;
	nfg_Vec P2Utils;
	foreach ( nfg_WtdInst wInst, wInsts )
	{
		// Map each instantiation (sample) of the net to the player's utilities
		// and save them to a list
		P1Utils.push_back( Util_function_P1(wInst.m_inst) );
		P2Utils.push_back( Util_function_P2(wInst.m_inst) );
	}
	// Output the utility values to a Matlab file for post-processing
	if_Matlab::Export_Vec_to_file ( "Output\\PD_vars.m", "P1Utils", P1Utils, true );
	if_Matlab::Export_Vec_to_file ( "Output\\PD_vars.m", "P2Utils", P2Utils, true );
}
// *************************************************************************
// *************************************************************************
void net_PrisonersDilemma::Analysis2 ( BayesNet& bnet )
{
	cout << "Running Analysis2..." << endl;
	if_Gambit::Export_to_efg_format( bnet, "Output\\PD.efg" );
}