
// *************************************************************************
//	LibNFG: Net-form game library
//
//	Author:		Ritchie Lee
//
//  Copyright 2011 United States Government as represented by the Administrator 
//  of the National Aeronautics and Space Administration.  All Rights Reserved.  
//  Licensed under the Apache License, Version 2.0 (the "License"); you may not 
//  use this file except in compliance with the License.  You may obtain a copy 
//  of the License at http://www.apache.org/licenses/LICENSE-2.0
//  Unless required by applicable law or agreed to in writing, software 
//  distributed under the License is distributed on an "AS IS" BASIS, WITHOUT 
//  WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the 
//  License for the specific language governing permissions and limitations 
//  under the License.
//
//	Revision Log
//
//	Username	Date		Description
//	-------------------------------------------------------
//	rlee		20110217	Created
//
// *************************************************************************

#include "net_student.h"

#include <boost/bind.hpp>
#include "bayesnet.h"
#include "cpd.h"
#include "cpd_tablebasic.h"
#include "alg_montecarlo.h"
#include "helper.h"
#include "if_bayesnettoolbox.h"

using namespace nfg;
using namespace std;
using namespace boost;

// *************************************************************************
// *************************************************************************
void net_Student::Run ( void )
{
	//*********************************************
	///	 Set up Bayes Net

	// Specify vertex names
	string vertexNames[] = 
	{	"D", "I", "G", "L", "S"	};

	// Define Edges of the graph
	nfg_Edge edgeArray[] = 
	{ 
		nfg_Edge(D,G), 
		nfg_Edge(I,G), nfg_Edge(I,S),
		nfg_Edge(G,L) 
	};

	//	Set up probability tables
	//	Probability table for D
	int			D_dims[]	= { 2 };
	double		D_probs[]	= { 0.6, 0.4 };

	//	Probability table for I
	int			I_dims[]	= { 2 };
	double		I_probs[]	= { 0.7, 0.3 };

	//	Probability table for G
	int			G_dims[]	= { 3, 2, 2 };
	double		G_probs[]	= { 0.3, 0.9, 0.05, 0.5, 
								0.4, 0.08, 0.25, 0.3,
								0.3, 0.02, 0.7, 0.2 };

	//	Probability table for L
	int			L_dims[]	= { 2, 3 };
	double		L_probs[]	= { 0.1, 0.4, 0.99,
								0.9, 0.6, 0.01 };

	//	Probability table for S
	int			S_dims[]	= { 2, 2 };
	double		S_probs[]	= { 0.95, 0.2,
								0.05, 0.8 };

	// Specify the space of each vertex
	nfg_Spaces spaces(N);
	spaces.at(D).push_back( Construct_discrete_space1d ( Vectorize(0,1) ));
	spaces.at(I).push_back( Construct_discrete_space1d ( Vectorize(0,1) ));
	spaces.at(G).push_back( Construct_discrete_space1d ( Vectorize(0,1,2) ));
	spaces.at(L).push_back( Construct_discrete_space1d ( Vectorize(0,1) ));
	spaces.at(S).push_back( Construct_discrete_space1d ( Vectorize(0,1) ));

	//	Make Bayes Net
	// Pass all specified values above to the bnet constructor
	BayesNet bnet( "Student", N, 
		nfg_Edges(edgeArray,edgeArray+sizeof(edgeArray)/sizeof(nfg_Edge)), 
		nfg_Dim(N,1),		// all sizes are 1
		spaces,				// possible values in each dimension
		nfg_Mask(N,false),	// all discrete values 
		nfg_eVertexTypes(N, CHANCE),	// all chance nodes
		nfg_Mask(N, false),			// no branch nodes
		nfg_Mask(N, false));		// no nodes with dno
	// Set vertex names
	bnet.Set_vertex_names ( vector<string>(vertexNames,vertexNames+N) );

	//	Create tabular cpds for each node
	cpd_TableBasic D_cpd ( nfg_Dim(D_dims,D_dims+sizeof(D_dims)/sizeof(int)), 
						nfg_Vec(D_probs,D_probs+sizeof(D_probs)/sizeof(double)));
	cpd_TableBasic I_cpd ( nfg_Dim(I_dims,I_dims+sizeof(I_dims)/sizeof(int)), 
						nfg_Vec(I_probs,I_probs+sizeof(I_probs)/sizeof(double)) );
	cpd_TableBasic G_cpd ( nfg_Dim(G_dims,G_dims+sizeof(G_dims)/sizeof(int)), 
						nfg_Vec(G_probs,G_probs+sizeof(G_probs)/sizeof(double)) );
	cpd_TableBasic L_cpd ( nfg_Dim(L_dims,L_dims+sizeof(L_dims)/sizeof(int)), 
						nfg_Vec(L_probs,L_probs+sizeof(L_probs)/sizeof(double)) );
	cpd_TableBasic S_cpd ( nfg_Dim(S_dims,S_dims+sizeof(S_dims)/sizeof(int)), 
						nfg_Vec(S_probs,S_probs+sizeof(S_probs)/sizeof(double)) );
	
	//	Link cpds to bnet using automatic linking
	bnet.Set_cpd(D, &D_cpd);
	bnet.Set_cpd(I, &I_cpd);
	bnet.Set_cpd(G, &G_cpd);
	bnet.Set_cpd(L, &L_cpd);
	bnet.Set_cpd(S, &S_cpd);

	// Error check
	bassert ( bnet.Check() );

	//**************************************************
	///	Run main program

	// Draw graph (to make sure the code understood your graph correctly)
	if_BayesNetToolbox::Draw_graph( "Output\\Stu_draw.m", bnet );

	// Suggest a topological ordering (as a test, it's already in order)
	bnet.Topological_sort();

	// Run custom analysis on bnet
	Analysis1( bnet );

	return;
}
// *************************************************************************
// *************************************************************************
void net_Student::Analysis1 ( BayesNet& bnet )
{
	// Declare inference engine
	alg_MonteCarlo engine;
	
	// Declare container for evidence
	nfg_OptInst evidence( N );

	// Test 1: set evidence to S and L
	evidence.at(S) = nfg_Vec(1, 1.0 ); 
	evidence.at(L) = nfg_Vec(1, 0.0 );

	// Estimate conditional probabilities of G using likelihood weighted sampling
	// Use 1000 samples
	cout << "P(G0|L1, S0): " 
		<< engine.LW_estimator( bnet, 1000, evidence, bind(&Indicator_iround, (int)G, 0, _1), true ) 
		<< endl;
	cout << "P(G1|L1, S0): " 
		<< engine.LW_estimator( bnet, 1000, evidence, bind(&Indicator_iround, (int)G, 1, _1), true ) 
		<< endl;
	cout << "P(G2|L1, S0): " 
		<< engine.LW_estimator( bnet, 1000, evidence, bind(&Indicator_iround, (int)G, 2, _1), true ) 
		<< endl;

	// Test 2: set evidence to D and S
	evidence.clear();
	evidence.resize( N );
	evidence.at(D) = nfg_Vec(1, 0.0 ); 
	evidence.at(S) = nfg_Vec(1, 1.0 );

	// Estimate conditional probabilities of L using likelihood weighted sampling
	// Use 1000 samples
	cout << "P(L0|D0, S1): " 
		<< engine.LW_estimator( bnet, 1000, evidence, bind(&Indicator_iround, (int)L, 0, _1), true ) 
		<< endl;
	cout << "P(L1|D0, S1): " 
		<< engine.LW_estimator( bnet, 1000, evidence, bind(&Indicator_iround, (int)L, 1, _1), true ) 
		<< endl;

	// Test 3: set evidence to D and L
	evidence.clear();
	evidence.resize( N );
	evidence.at(D) = nfg_Vec(1, 0.0 ); 
	evidence.at(L) = nfg_Vec(1, 0.0 );

	// Estimate conditional probabilities of S using likelihood weighted sampling
	// Use 1000 samples
	cout << "P(S0|D0, L0): " 
		<< engine.LW_estimator( bnet, 1000, evidence, bind(&Indicator_iround, (int)S, 0, _1), true ) 
		<< endl;
	cout << "P(S1|D0, L0): " 
		<< engine.LW_estimator( bnet, 1000, evidence, bind(&Indicator_iround, (int)S, 1, _1), true ) 
		<< endl;

	return;
}

