
// *************************************************************************
//	LibNFG: Net-form game library
//
//	Author:		Ritchie Lee
//
//  Copyright 2011 United States Government as represented by the Administrator 
//  of the National Aeronautics and Space Administration.  All Rights Reserved.  
//  Licensed under the Apache License, Version 2.0 (the "License"); you may not 
//  use this file except in compliance with the License.  You may obtain a copy 
//  of the License at http://www.apache.org/licenses/LICENSE-2.0
//  Unless required by applicable law or agreed to in writing, software 
//  distributed under the License is distributed on an "AS IS" BASIS, WITHOUT 
//  WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the 
//  License for the specific language governing permissions and limitations 
//  under the License.
//
//	Revision Log
//
//	Username	Date		Description
//	-------------------------------------------------------
//	rlee		20110721	Created
//	rlee		20110726	Renamed class to PointMassKinematics
//
// *************************************************************************

#ifndef INCLUDED_POINTMASSKINEMATICS_H
#define INCLUDED_POINTMASSKINEMATICS_H

#include <vector>

using namespace std;

/**
 * Point mass physics simulation in 3D.  State is described by
 *	3D position and 3D velocity.  Use Euler integration for propagating
 *	forward in discrete time.  Take accelerations at each time step
 *	to update the 6-D state.
 */
class PointMassKinematics
{
public:

	/**	Constructor */
	PointMassKinematics( void );
	PointMassKinematics( double x0, double y0, double z0, 
							double vx0, double vy0, double vz0 );

	/**	Destructor */
	virtual ~ PointMassKinematics();

	/// Variables

	///	Methods
	/**
	 * Set_params: Set parameters for needed physics simulation
	 *	dt - time step size for Euler integration step
	 */
	void Set_params ( double dt );
	/**
	 * Set_state: Override and manually set the state.
	 *	x, y, z - x, y, and z position of point mass
	 *	vx, vy, vz - x, y, and z velocity of point mass
	 */
	void Set_state( double x, double y, double z,
							double vx, double vy, double vz );
	/**
	 * Step: Perform an update step advancing the state one time step
	 *		using the given acceleration and according to equations of motion.
	 *	ax, ay, az - x, y, and z acceleration components of point mass
	 */
	void Step( double ax, double ay, double az );
	/**
	 * Get_x, Get_y, Get_z: Get the current x, y, or z position of particle
	 *	returns: Current x, y, or z position.
	 */
	double Get_x( void ) { return m_x; };
	double Get_y( void ) { return m_y; };
	double Get_z( void ) { return m_z; };
	/**
	 * Get_vx, Get_vy, Get_vz: Get the current x, y, or z velocity of particle
	 *	returns: Current x, y, or z velocity.
	 */
	double Get_vx( void ) { return m_vx; };
	double Get_vy( void ) { return m_vy; };
	double Get_vz( void ) { return m_vz; };
	/**
	 * Get_position_3d: Get the current x, y, or z position of particle as
	 *		a vector of doubles
	 *	returns: Current position in 3D
	 */
	vector<double> Get_position_3d( void );
	/**
	 * Get_velocity_3d: Get the current x, y, or z velocity of particle as
	 *		a vector of doubles
	 *	returns: Current velocity in 3D
	 */
	vector<double> Get_velocity_3d( void );

	// Debug
	static void		Unit_test( void );

private:

	/// Variables
	double		m_x;		// x position of particle
	double		m_y;		// y position of particle
	double		m_z;		// z position of particle
	double		m_vx;		// x velocity of particle
	double		m_vy;		// y velocity of particle
	double		m_vz;		// z velocity of particle
	double		m_dt;		// time step size

	/// Methods

};

#endif

