
// *************************************************************************
//	LibNFG: Net-form game library
//
//	Author:		Ritchie Lee
//
//  Copyright 2011 United States Government as represented by the Administrator 
//  of the National Aeronautics and Space Administration.  All Rights Reserved.  
//  Licensed under the Apache License, Version 2.0 (the "License"); you may not 
//  use this file except in compliance with the License.  You may obtain a copy 
//  of the License at http://www.apache.org/licenses/LICENSE-2.0
//  Unless required by applicable law or agreed to in writing, software 
//  distributed under the License is distributed on an "AS IS" BASIS, WITHOUT 
//  WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the 
//  License for the specific language governing permissions and limitations 
//  under the License.
//
//	Revision Log
//
//	Username	Date		Description
//	-------------------------------------------------------
//	rlee		20110404	Created
//
// *************************************************************************

#ifndef INCLUDED_TESTCASES_H
#define INCLUDED_TESTCASES_H

#include <vector>
#include <list>
#include <string>
#include <boost/any.hpp>

using namespace std;
using namespace boost;

typedef list<any>				nfg_tc_Channel;
typedef vector<any>				nfg_tc_TestCase;
typedef list<nfg_tc_Channel>	nfg_tc_Channels;

/**
 * TestCases: Automatic testcase generator/enumerator.
 *	Input basic data and use the different iterators to generate/enumerate.
 *	Organized by channels of parameters.
 *	Ch1: ch1_param1, ch1_param2, ...
 *	Ch2: ch2_param1, ch2_param2, ...
 *	Output will be a vector with a param from each channel.
 */
class TestCases
{
public:

	/**	Constructor */
	TestCases( void ) {};

	/**	Destructor */
	virtual ~ TestCases(){};

	/// Variables

	///	Methods
	/**
	 * Add_channel_to_end: Add a channel to the end of the channel list
	 *	channel - channel to be added
	 *	name - channel name (optional)
	 */
	void	Add_channel_to_end ( const nfg_tc_Channel& channel, const string name = "" );
	/**
	 * Get_num_channels: Get number of channels
	 *	return: number of channels
	 */
	int		Get_num_channels ( void ) { return m_data.size(); };
	/**
	 * Delete_last_channel: Delete last channel
	 */
	void	Delete_last_channel ( void );
	/**
	 * Get_channel_name: Get name of channel
	 *	id - channel id of interest
	 *	return: name of channel
	 */
	string	Get_channel_name ( int id ) { return m_channelNames.at(id); };

	//	Debug
	static void		Unit_test( void );	// test combo and straight iterators

protected:

	/// Variables
	nfg_tc_Channels		m_data;			// Main container object
	vector<string>		m_channelNames;	// List of channel names

	/// Methods

public:
	/**
	 * ComboIterator: Enumerates all combinations of channel parameters
	 */
	class ComboIterator
	{

	public:
		ComboIterator ( void ) {};
		/**
		 * ComboIterator: Constructor.  Associate with TestCase and set
		 *		to the beginning
		 *	src - testcase object to point to
		 */
		ComboIterator ( TestCases& src ) { operator=(src); };
		virtual ~ComboIterator() {};
		/**
		 * operator=: Associate iterator with a TestCase object and initialize
		 *		to the beginning
		 *	src - testcase object to point to
		 */
		void operator=(TestCases& src );
		/**
		 * operator++: Increment iterator to the next element
		 */
		void operator++( void );
		/**
		 * operator(): Return element at current iterator position
		 */
		nfg_tc_TestCase operator()( void );

		
	private:
		vector<nfg_tc_Channel::iterator>		m_index;		// current position in each channel
		vector<nfg_tc_Channel::iterator>		m_begin;		// start position of each channel
		vector<nfg_tc_Channel::iterator>		m_end;			// end position of each channel

	};

	/**
	 * ParallelIterator: Iterates in parallel through channel parameters
	 *	i.e., iterates by advancing all channels by 1
	 */
	class ParallelIterator
	{

	public:
		ParallelIterator ( void ) {};
		/**
		 * ParallelIterator: Constructor.  Associate with TestCase and set
		 *		to the beginning
		 *	src - testcase object to point to
		 */
		ParallelIterator ( TestCases& src ) { operator=(src); };
		virtual ~ParallelIterator() {};
		/**
		 * operator=: Associate iterator with a TestCase object and initialize
		 *		to the beginning
		 *	src - testcase object to point to
		 */
		void operator=(TestCases& src );
		/**
		 * operator++: Increment iterator to the next element
		 */
		void operator++( void );
		/**
		 * operator(): Return element at current iterator position
		 */
		nfg_tc_TestCase operator()( void );

		
	private:
		vector<nfg_tc_Channel::iterator>		m_index;	// current position in each channel
		vector<nfg_tc_Channel::iterator>		m_begin;	// start position of each channel
		vector<nfg_tc_Channel::iterator>		m_end;		// end position of each channel

	};

};
#endif

