
// *************************************************************************
//	LibNFG: Net-form game library
//
//	Author:		Ritchie Lee
//
//  Copyright 2011 United States Government as represented by the Administrator 
//  of the National Aeronautics and Space Administration.  All Rights Reserved.  
//  Licensed under the Apache License, Version 2.0 (the "License"); you may not 
//  use this file except in compliance with the License.  You may obtain a copy 
//  of the License at http://www.apache.org/licenses/LICENSE-2.0
//  Unless required by applicable law or agreed to in writing, software 
//  distributed under the License is distributed on an "AS IS" BASIS, WITHOUT 
//  WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the 
//  License for the specific language governing permissions and limitations 
//  under the License.
//
//	Revision Log
//
//	Username	Date		Description
//	-------------------------------------------------------
//	rlee		20101025	Created
//
// *************************************************************************

#ifndef INCLUDED_TYPES_H
#define INCLUDED_TYPES_H

#include <vector>
#include <list>
#include <boost/assert.hpp>
#include <boost/accumulators/accumulators.hpp>
#include <boost/accumulators/statistics/stats.hpp>
#include <boost/accumulators/statistics/sum.hpp>
#include <boost/accumulators/statistics/weighted_sum.hpp>
#include <boost/foreach.hpp>
#include <boost/function.hpp>
#include <boost/optional.hpp>
#include <boost/serialization/serialization.hpp>

using namespace std;
using namespace boost;
using namespace boost::accumulators;

#define foreach				BOOST_FOREACH	// macro from boost. not the same as for_each
#define bassert				BOOST_ASSERT	// 


//	Object types
typedef vector<double>						nfg_Vec;			//	Single random vector (basic unit)
typedef vector<nfg_Vec>						nfg_Vecs;			//	List of vectors
typedef vector<nfg_Vec>						nfg_Inst;			//	List of vectors. size of bnet, an instantiation
typedef list<nfg_Inst>						nfg_Insts;			//	List of instantiations
typedef vector< optional<nfg_Vec> >			nfg_OptInst;		//	List of optional vectors. size of bnet, an optional instantiation
typedef vector<bool>						nfg_Mask;			//	Boolean for masking purposes (e.g. defining sets)
typedef vector<nfg_Mask>					nfg_Masks;			//	List of masks (e.g. one mask for each node)
typedef vector<unsigned int>				nfg_Dim;			//	Dimensions (i.e., tuple)
typedef vector<nfg_Dim>						nfg_Dims;			//	A vector of Dimensions

// Weighted nfg_Vec
struct nfg_WtdVec
{
	nfg_WtdVec(){};	// Default constructor
	// Assigning constructor, for convenience
	nfg_WtdVec( const nfg_Vec& vec, const double weight )
	{
		m_vec = vec;
		m_weight = weight;
	};

	nfg_Vec			m_vec;
	double			m_weight;
};

// Weighted nfg_Inst
struct nfg_WtdInst
{
	nfg_WtdInst(){};	// Default constructor
	// Assigning constructor, for convenience
	nfg_WtdInst( const nfg_Inst& inst, 	const double weight )
	{
		m_inst = inst;
		m_weight = weight;
	};
	
	friend class boost::serialization::access;

    template<class Archive>
    void serialize(Archive & ar, const unsigned int version)
    {
        ar & m_inst;
        ar & m_weight;
    }

	nfg_Inst		m_inst;
	double			m_weight;
};

// Weighted nfg_OptInst
struct nfg_OptWtdInst
{
	nfg_OptWtdInst(){};		// Default constructor
	// Assigning constructor, for convenience
	nfg_OptWtdInst( const nfg_OptInst& optInst, const double weight )
	{
		m_optInst = optInst;
		m_weight = weight;
	};

	nfg_OptInst		m_optInst;
	double			m_weight;
};

typedef vector<nfg_WtdInst>									nfg_WtdInsts;		// weighted nfg_Insts
typedef vector<nfg_OptWtdInst>								nfg_OptWtdInsts;	// optional nfg_WtdInsts
typedef accumulator_set< double, stats<tag::sum_of_weights,tag::weighted_sum>, double >	// statistical accumulator
															nfg_AccumWtdSum;	// for weighted sums

typedef function<double (const nfg_Inst& instVecs)>			nfg_UtilFunc;		// Utility function object
typedef function<nfg_Vec (const nfg_Vecs& paVecs)>			nfg_SampleFunc;		// Sample function object (e.g. for sample_node, or sample_move)
typedef function<double (const nfg_Vec& outVec, const nfg_Vecs& paVecs)>		// Get_prob function object
															nfg_ProbFunc;
typedef vector<nfg_SampleFunc>								nfg_SampleFuncs;	// vector of nfg_Sample
typedef function<bool (const nfg_OptInst& optInstVecs)>		nfg_DnoFunc;		// function for determining dno
typedef vector<nfg_DnoFunc>									nfg_DnoFuncs;		// vector of nfg_DnoFunc
typedef pair<double, double>								nfg_Range;			// min/max range using doubles
typedef vector<nfg_Range>									nfg_Space1D;		// Represents a single-dimensional space
typedef vector<nfg_Space1D>									nfg_Space;			// Represents a multi-dimensional space
typedef vector<nfg_Space>									nfg_Spaces;			// Vector of spaces 
typedef function<nfg_Vec (const nfg_OptWtdInst& partWtdInst)>	
															nfg_PropFunc;		// sample proposal distribution
typedef function<nfg_Vec (void)>							nfg_PropFuncDirect;	// Direct call to sample proposal distribution, no inputs needed
typedef vector<nfg_PropFunc>								nfg_PropFuncInst;		// vector (size of bnet) of proposal distribution functions
typedef vector<nfg_PropFuncInst>							nfg_PropFuncInsts;		// vector of nfg_PropFuncInst
typedef function<double (const nfg_Vec& outVec, const nfg_OptWtdInst& partWtdInst)>	
															nfg_PropProbFunc;		// Get_probability for proposal distr
typedef function<double (const nfg_Vec& outVec)>			nfg_PropProbFuncDirect;	// Direct call to proposal Get_probability, already conditioned
typedef vector<nfg_PropProbFunc>							nfg_PropProbFuncInst;	// vector (size of bnet) of proposal Get_probability functions
typedef vector<nfg_PropProbFuncInst>						nfg_PropProbFuncInsts;	// vector of nfg_PropProbFuncInst

typedef nfg_Inst											nfg_Slice;
typedef vector<nfg_Slice>									nfg_Episode;
typedef vector<nfg_Episode>									nfg_Episodes;

#endif

